#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Yannick Dayer <yannick.dayer@idiap.ch>
# Mon 14 Sep 2020 17:00:41 UTC+02

"""Tests for the bob.bio.base.script.gen module

The gen module generates synthetic scores and saves them to a file for
demonstration and test purpose.
"""

import logging
import os

import numpy

from click.testing import CliRunner

from bob.bio.base.script.gen import gen, gen_score_distr
from bob.io.base.testing_utils import assert_click_runner_result

logger = logging.getLogger(__name__)
logger.setLevel(
    "DEBUG"
)  # If NOTSET (default), will be changed to ERROR at CliRunner.invoke


def gen_case(
    n_subjects,
    n_probes_per_subject,
    n_unknown_subjects,
    n_pos=None,
    n_neg=None,
    n_unk=None,
):
    """Tests one case of the gen command"""
    n_pos = n_subjects * n_probes_per_subject if n_pos is None else n_pos
    n_neg = (
        n_subjects * (n_subjects - 1) * n_probes_per_subject
        if n_neg is None
        else n_neg
    )
    n_unk = (
        n_unknown_subjects * n_subjects * n_probes_per_subject
        if n_unk is None
        else n_unk
    )

    runner = CliRunner()
    with runner.isolated_filesystem():
        temp_path = "./gen_test_temp_dir/"

        logger.info("Calling 'gen' with a specific amount of scores.")
        result = runner.invoke(
            gen,
            args=[
                "-mm",
                "10",
                "-mnm",
                "-10",
                "-sp",
                "1",
                "-sn",
                "1",
                "-p",
                f"{n_probes_per_subject}",
                "-s",
                f"{n_subjects}",
                "-u",
                f"{n_unknown_subjects}",
                "-f",
                "--n-pos",
                f"{n_pos}",
                "--n-neg",
                f"{n_neg}",
                "--n-unk",
                f"{n_unk}",
                f"{temp_path}",
            ],
        )
        assert_click_runner_result(result)
        assert os.path.exists(
            os.path.join(temp_path, "scores-dev.csv")
        ), "dev scores file not created."
        line_count = 0
        with open(os.path.join(temp_path, "scores-dev.csv")) as f:
            for _ in f:
                line_count += 1
        assert line_count == 1 + n_pos + n_neg + n_unk
        assert os.path.exists(
            os.path.join(temp_path, "scores-eval.csv")
        ), "eval scores file not created."
        line_count = 0
        with open(os.path.join(temp_path, "scores-eval.csv")) as f:
            for _ in f:
                line_count += 1
        assert line_count == 1 + n_pos + n_neg + n_unk


def test_gen():
    """Tests that the main gen command works as expected"""
    # Define a click runner to invoke click commands

    logger.info("Calling 'gen' with a specific amount of scores.")
    gen_case(
        n_subjects=5,
        n_probes_per_subject=5,
        n_unknown_subjects=2,
        n_pos=10,
        n_neg=60,
        n_unk=20,
    )

    logger.info("Calling 'gen' without a specific amount.")
    gen_case(
        n_subjects=5,
        n_probes_per_subject=5,
        n_unknown_subjects=2,
    )

    logger.info("Calling 'gen' without unknown subjects.")
    gen_case(
        n_subjects=5,
        n_probes_per_subject=2,
        n_unknown_subjects=0,
    )

    logger.info("Calling 'gen' with no subjects.")
    gen_case(
        n_subjects=0,
        n_probes_per_subject=2,
        n_unknown_subjects=0,
    )

    logger.info("Calling 'gen' with no probes.")
    gen_case(
        n_subjects=5,
        n_probes_per_subject=0,
        n_unknown_subjects=2,
    )

    logger.info("Calling 'gen' with only unknowns.")
    gen_case(
        n_subjects=5,
        n_probes_per_subject=0,
        n_unknown_subjects=2,
    )


def test_gen_score_dist():
    """Tests that the scores generation works as expected"""
    neg, pos = gen_score_distr(
        mean_neg=-10,
        mean_pos=10,
        sigma_neg=1,
        sigma_pos=1,
        n_neg=20,
        n_pos=20,
        seed=0,
    )
    assert (
        len(neg) == 20
    ), f"Incorrect number of negative scores generated ({len(neg)})"
    assert (
        len(pos) == 20
    ), f"Incorrect number of positive scores generated ({len(pos)})"
    assert all(
        [isinstance(s, (numpy.floating, float)) for s in neg]
    ), "A score was not a float"
    assert all(
        [isinstance(s, (numpy.floating, float)) for s in pos]
    ), "A score was not a float"
    expected_neg = numpy.array(
        [
            -8.23594765,
            -9.59984279,
            -9.02126202,
            -7.7591068,
            -8.13244201,
            -10.97727788,
            -9.04991158,
            -10.15135721,
            -10.10321885,
            -9.5894015,
            -9.85595643,
            -8.54572649,
            -9.23896227,
            -9.87832498,
            -9.55613677,
            -9.66632567,
            -8.50592093,
            -10.20515826,
            -9.6869323,
            -10.85409574,
        ]
    )
    expected_pos = numpy.array(
        [
            7.44701018,
            10.6536186,
            10.8644362,
            9.25783498,
            12.26975462,
            8.54563433,
            10.04575852,
            9.81281615,
            11.53277921,
            11.46935877,
            10.15494743,
            10.37816252,
            9.11221425,
            8.01920353,
            9.65208785,
            10.15634897,
            11.23029068,
            11.20237985,
            9.61267318,
            9.69769725,
        ]
    )
    assert numpy.allclose(neg, expected_neg), "Unexpected score generated"
    assert numpy.allclose(pos, expected_pos), "Unexpected score generated"



import logging
import os
import tarfile
import tempfile

import h5py
import numpy

logger = logging.getLogger("bob.bio.base")

import bob.io.base


def filter_missing_files(
    file_names, split_by_client=False, allow_missing_files=True
):
    """This function filters out files that do not exist, but only if ``allow_missing_files`` is set to ``True``, otherwise the list of ``file_names`` is returned unaltered."""

    if not allow_missing_files:
        return file_names

    if split_by_client:
        # filter out missing files and empty clients
        existing_files = [
            [f for f in client_files if os.path.exists(f)]
            for client_files in file_names
        ]
        existing_files = [
            client_files for client_files in existing_files if client_files
        ]
    else:
        # filter out missing files
        existing_files = [f for f in file_names if os.path.exists(f)]
    return existing_files


def filter_none(data, split_by_client=False):
    """This function filters out ``None`` values from the given list (or list of lists, when ``split_by_client`` is enabled)."""

    if split_by_client:
        # filter out missing files and empty clients
        existing_data = [
            [d for d in client_data if d is not None] for client_data in data
        ]
        existing_data = [
            client_data for client_data in existing_data if client_data
        ]
    else:
        # filter out missing files
        existing_data = [d for d in data if d is not None]
    return existing_data


def check_file(filename, force, expected_file_size=1):
    """Checks if the file with the given ``filename`` exists and has size greater or equal to ``expected_file_size``.
    If the file is to small, **or** if the ``force`` option is set to ``True``, the file is removed.
    This function returns ``True`` is the file exists (and has not been removed), otherwise ``False``
    """
    if os.path.exists(filename):
        if force or os.path.getsize(filename) < expected_file_size:
            logger.debug("  .. Removing old file '%s'.", filename)
            os.remove(filename)
            return False
        else:
            return True
    return False


def read_original_data(biofile, directory, extension):
    """This function reads the original data using the given ``biofile`` instance.
    It simply calls ``load(directory, extension)`` from :py:class:`bob.bio.base.database.BioFile` or one of its derivatives.

    Parameters
    ----------
    biofile : :py:class:`bob.bio.base.database.BioFile` or one of its derivatives
        The file to read the original data.
    directory : str
        The base directory of the database.
    extension : str or ``None``
        The extension of the original data.
        Might be ``None`` if the ``biofile`` itself has the extension stored.

    Returns
    -------
    object
        Whatver ``biofile.load`` returns; usually a :py:class:`numpy.ndarray`
    """
    return biofile.load(directory, extension)


def load(file):
    """Loads data from file. The given file might be an HDF5 file open for reading or a string."""
    if isinstance(file, h5py.File):
        return numpy.array(file["array"])
    else:
        return bob.io.base.load(file)


def save(data, file, compression=0):
    """Saves the data to file using HDF5. The given file might be an HDF5 file open for writing, or a string.
    If the given data contains a ``save`` method, this method is called with the given HDF5 file.
    Otherwise the data is written to the HDF5 file using the given compression.
    """
    f = file if isinstance(file, h5py.File) else h5py.File(file, "w")
    if hasattr(data, "save"):
        data.save(f)
    else:
        f["array"] = data


def open_compressed(filename, open_flag="r", compression_type="bz2"):
    """Opens a compressed HDF5File with the given opening flags.
    For the 'r' flag, the given compressed file will be extracted to a local space.
    For 'w', an empty HDF5File is created.
    In any case, the opened HDF5File is returned, which needs to be closed using the close_compressed() function.
    """
    # create temporary HDF5 file name
    hdf5_file_name = tempfile.mkstemp(".hdf5", "bob_")[1]

    if open_flag == "r":
        # extract the HDF5 file from the given file name into a temporary file name
        tar = tarfile.open(filename, mode="r:" + compression_type)
        memory_file = tar.extractfile(tar.next())
        real_file = open(hdf5_file_name, "wb")
        real_file.write(memory_file.read())
        del memory_file
        real_file.close()
        tar.close()

    return h5py.File(hdf5_file_name, open_flag)


def close_compressed(
    filename, hdf5_file, compression_type="bz2", create_link=False
):
    """Closes the compressed hdf5_file that was opened with open_compressed.
    When the file was opened for writing (using the 'w' flag in open_compressed), the created HDF5 file is compressed into the given file name.
    To be able to read the data using the real tools, a link with the correct extension might is created, when create_link is set to True.
    """
    hdf5_file_name = hdf5_file.filename
    is_writable = hdf5_file.writable
    hdf5_file.close()

    if is_writable:
        # create compressed tar file
        tar = tarfile.open(filename, mode="w:" + compression_type)
        tar.add(hdf5_file_name, os.path.basename(filename))
        tar.close()

    if create_link:
        extension = {"": ".tar", "bz2": ".tar.bz2", "gz": "tar.gz"}[
            compression_type
        ]
        link_file = filename + extension
        if not os.path.exists(link_file):
            os.symlink(os.path.basename(filename), link_file)

    # clean up locally generated files
    os.remove(hdf5_file_name)


def load_compressed(filename, compression_type="bz2"):
    """Extracts the data to a temporary HDF5 file using HDF5 and reads its contents.
    Note that, though the file name is .hdf5, it contains compressed data!
    Accepted compression types are 'gz', 'bz2', ''"""
    # read from compressed HDF5
    hdf5 = open_compressed(filename, "r")
    data = numpy.array(hdf5["array"])
    close_compressed(filename, hdf5)

    return data


def save_compressed(data, filename, compression_type="bz2", create_link=False):
    """Saves the data to a temporary file using HDF5.
    Afterwards, the file is compressed using the given compression method and saved using the given file name.
    Note that, though the file name will be .hdf5, it will contain compressed data!
    Accepted compression types are 'gz', 'bz2', ''"""
    # write to compressed HDF5 file
    hdf5 = open_compressed(filename, "w")
    save(data, hdf5)
    close_compressed(filename, hdf5, compression_type, create_link)



#!/usr/bin/env python
# Ivana Chingovska <ivana.chingovska@idiap.ch>
# Fri Dec  7 12:33:37 CET 2012
"""Utility functions for computation of EPSC curve and related measurement"""

import numpy

from bob.measure import farfrr


def calc_pass_rate(threshold, attacks):
    """Calculates the rate of successful spoofing attacks

    Parameters
    ----------
    threshold :
      the threshold used for classification
    scores :
      numpy with the scores of the spoofing attacks

    Returns
    -------
    float
      rate of successful spoofing attacks
    """
    return (attacks >= threshold).mean()


def weighted_neg_error_rate_criteria(
    data, weight, thres, beta=0.5, criteria="eer"
):
    """Given the single value for the weight parameter balancing between
    impostors and spoofing attacks and a threshold, calculates the error rates
    and their relationship depending on the criteria (difference in case of
    'eer', hter in case of 'min-hter' criteria)
    Keyword parameters:

      - data - the development data used to determine the threshold. List on 4
      numpy.arrays containing: negatives (licit), positives (licit),
      negatives (spoof), positives (spoof)
      - weight - the weight parameter balancing between impostors and spoofing
      attacks
      - thres - the given threshold
      - beta - the weight parameter balancing between real accesses and all the
      negative samples (impostors and spoofing attacks). Note that this
      parameter will be overridden and not considered if the selected criteria
      is 'min-hter'.
      - criteria - 'eer', 'wer' or 'min-hter' criteria for decision threshold
    """

    licit_neg = data[0]
    licit_pos = data[1]
    spoof_neg = data[2]
    spoof_pos = data[3]  # unpacking the data
    farfrr_licit = farfrr(licit_neg, licit_pos, thres)
    farfrr_spoof = farfrr(spoof_neg, spoof_pos, thres)

    frr = farfrr_licit[1]  # farfrr_spoof[1] should have the same value
    far_i = farfrr_licit[0]
    far_s = farfrr_spoof[0]

    far_w = (1 - weight) * far_i + weight * far_s

    if criteria == "eer":
        if beta == 0.5:
            return abs(far_w - frr)
        else:
            # return abs(far_w - frr)
            return abs((1 - beta) * frr - beta * far_w)

    elif criteria == "min-hter":
        return (far_w + frr) / 2

    else:
        return (1 - beta) * frr + beta * far_w


def epsc_weights(licit_neg, licit_pos, spoof_neg, spoof_pos, points=100):
    """Returns the weights for EPSC

    Keyword arguments:

      - points - number of points to calculate EPSC
    """
    step_size = 1 / float(points)
    weights = numpy.array([(i * step_size) for i in range(points + 1)])
    return weights


def recursive_thr_search(
    data, span_min, span_max, weight, beta=0.5, criteria="eer"
):
    """Recursive search for the optimal threshold given a criteria. It
      evaluates the full range of thresholds at 100 points, and computes the one
      which optimizes the threshold. In the next search iteration, it examines
      the region around the point that optimizes the threshold. The procedure
      stops when the search range is smaller then 1e-10.

    Keyword arguments:
      - data - the development data used to determine the threshold. List on 4
      numpy.arrays containing: negatives (licit), positives (licit), negatives
      (spoof), positives (spoof)
      - span_min - the minimum of the search range
      - span_max - the maximum of the search range
      - weight - the weight parameter balancing between impostors and spoofing
      attacks
      - beta - the weight parameter balancing between real accesses and all the
      negative samples (impostors and spoofing attacks). Note that methods called
      within this function will override this parameter and not considered if the
      selected criteria is 'min-hter'.
      - criteria - the decision threshold criteria ('eer' for EER, 'wer' for
      Minimum WER or 'min-hter' for Minimum HTER criteria).
    """

    quit_thr = 1e-10
    steps = 100
    if abs((span_max - span_min) / span_max) < quit_thr:
        return span_max  # or span_min, it doesn't matter
    else:
        step_size = (span_max - span_min) / steps
        thresholds = numpy.array(
            [(i * step_size) + span_min for i in range(steps + 1)]
        )
        weighted_error_rates = numpy.array(
            [
                weighted_neg_error_rate_criteria(
                    data, weight, thr, beta, criteria
                )
                for thr in thresholds
            ]
        )
        selected_thres = thresholds[
            numpy.where(weighted_error_rates == min(weighted_error_rates))
        ]  # all the thresholds which have minimum weighted error rate
        thr = selected_thres[
            int(selected_thres.size / 2)
        ]  # choose the centrally positioned threshold
        return recursive_thr_search(
            data, thr - step_size, thr + step_size, weight, beta, criteria
        )


def weighted_negatives_threshold(
    licit_neg, licit_pos, spoof_neg, spoof_pos, weight, beta=0.5, criteria="eer"
):
    """Calculates the threshold for achieving the given criteria between the
      FAR_w and the FRR, given the single value for the weight parameter
      balancing between impostors and spoofing attacks and a single value for the
      parameter beta balancing between the real accesses and the negatives
      (impostors and spoofing attacks)

    Keyword parameters:
      - licit_neg - numpy.array of scores for the negatives (licit scenario)
      - licit_pos - numpy.array of scores for the positives (licit scenario)
      - spoof_neg - numpy.array of scores for the negatives (spoof scenario)
      - spoof_pos - numpy.array of scores for the positives (spoof scenario)
      - weight - the weight parameter balancing between impostors and spoofing
      attacks
      - beta - the weight parameter balancing between real accesses and all the
      negative samples (impostors and spoofing attacks). Note that methods called
      within this function will override this parameter and not considered if the
      selected criteria is 'min-hter'.
      - criteria - the decision threshold criteria ('eer' for EER, 'wer' for
      Minimum WER or 'min-hter' for Minimum HTER criteria).
    """
    span_min = min(
        numpy.append(licit_neg, spoof_neg)
    )  # the min of the span where we will search for the threshold
    span_max = max(
        numpy.append(licit_pos, spoof_pos)
    )  # the max of the span where we will search for the threshold
    data = (
        licit_neg,
        licit_pos,
        spoof_neg,
        spoof_pos,
    )  # pack the data into a single list
    return recursive_thr_search(
        data, span_min, span_max, weight, beta, criteria
    )


def epsc_thresholds(
    licit_neg,
    licit_pos,
    spoof_neg,
    spoof_pos,
    points=100,
    criteria="eer",
    omega=None,
    beta=None,
):
    """Calculates the optimal thresholds for EPSC, for a range of the weight
      parameter balancing between impostors and spoofing attacks, and for a range
      of the beta parameter balancing between real accesses and all the negatives
      (impostors and spoofing attacks)

    Keyword arguments:

      - licit_neg - numpy.array of scores for the negatives (licit scenario)
      - licit_pos - numpy.array of scores for the positives (licit scenario)
      - spoof_neg - numpy.array of scores for the negatives (spoof scenario)
      - spoof_pos - numpy.array of scores for the positives (spoof scenario)
      - points - number of points to calculate EPSC
      - criteria - the decision threshold criteria ('eer', 'wer' or 'min-hter')
      - omega - the value of the parameter omega, balancing between impostors and
      spoofing attacks. If None, it is going to span the full range [0,1].
      Otherwise, can be set to a fixed value or a list of values.
      - beta - the value of the parameter beta, balancing between real accesses
      and all the negatives (zero-effort impostors and spoofing attacks). If
      None, it is going to span the full range [0,1]. Otherwise, can be set to a
      fixed value or a list of values.

    """
    step_size = 1 / float(points)

    if omega is None:
        omega = numpy.array([(i * step_size) for i in range(points + 1)])
    elif (
        not isinstance(omega, list)
        and not isinstance(omega, tuple)
        and not isinstance(omega, numpy.ndarray)
    ):
        omega = numpy.array([omega])
    else:
        omega = numpy.array(omega)

    if beta is None:
        beta = numpy.array([(i * step_size) for i in range(points + 1)])
    elif (
        not isinstance(beta, list)
        and not isinstance(beta, tuple)
        and not isinstance(beta, numpy.ndarray)
    ):
        beta = numpy.array([beta])
    else:
        beta = numpy.array(beta)

    thresholds = numpy.ndarray([beta.size, omega.size], "float64")
    for bindex, b in enumerate(beta):
        thresholds[bindex, :] = numpy.array(
            [
                weighted_negatives_threshold(
                    licit_neg,
                    licit_pos,
                    spoof_neg,
                    spoof_pos,
                    w,
                    b,
                    criteria=criteria,
                )
                for w in omega
            ],
            "float64",
        )

    return omega, beta, thresholds


def weighted_err(error_1, error_2, weight):
    """Calculates the weighted error rate between the two input parameters

    Keyword arguments:
      - error_1 - the first input error rate (FAR for zero effort impostors
      usually)
      - error_2 - the second input error rate (SFAR)
      - weight - the given weight
    """
    return (1 - weight) * error_1 + weight * error_2


def error_rates_at_weight(
    licit_neg, licit_pos, spoof_neg, spoof_pos, omega, threshold, beta=0.5
):
    """Calculates several error rates: FRR, FAR (zero-effort impostors), SFAR,
        FAR_w, HTER_w for a given value of w. It returns the calculated threshold
        as a last argument

      Keyword arguments:

        - licit_neg - numpy.array of scores for the negatives (licit scenario)
        - licit_pos - numpy.array of scores for the positives (licit scenario)
        - spoof_neg - numpy.array of scores for the negatives (spoof scenario)
        - spoof_pos - numpy.array of scores for the positives (spoof scenario)
        - threshold - the given threshold
        - omega - the omega parameter balancing between impostors and spoofing
        attacks
        - beta - the weight parameter balancing between real accesses and all the

    negative samples (impostors and spoofing attacks).
    """

    farfrr_licit = farfrr(
        licit_neg, licit_pos, threshold
    )  # calculate test frr @ threshold (licit scenario)
    farfrr_spoof = farfrr(
        spoof_neg, spoof_pos, threshold
    )  # calculate test frr @ threshold (spoof scenario)

    # we can take this value from farfrr_spoof as well, it doesn't matter
    frr = farfrr_licit[1]
    far = farfrr_licit[0]
    sfar = farfrr_spoof[0]

    far_w = weighted_err(far, sfar, omega)
    hter_w = (far_w + frr) / 2
    wer_wb = weighted_err(frr, far_w, beta)

    return (frr, far, sfar, far_w, wer_wb, hter_w, threshold)


def epsc_error_rates(
    licit_neg, licit_pos, spoof_neg, spoof_pos, thresholds, omega, beta
):
    """Calculates several error rates: FAR_w and WER_wb for the given weights
    (omega and beta) and thresholds (the thresholds need to be computed first
    using the method: epsc_thresholds() before passing to this method)

    Parameters
    ----------
    licit_neg : array_like
        array of scores for the negatives (licit scenario)
    licit_pos : array_like
        array of scores for the positives (licit scenario)
    spoof_neg : array_like
        array of scores for the negatives (spoof scenario)
    spoof_pos : array_like
        array of scores for the positives (spoof scenario)
    thresholds : array_like
        ndarray with threshold values
    omega : array_like
        array of the omega parameter balancing between impostors
        and spoofing attacks
    beta : array_like
        array of the beta parameter balancing between real accesses
        and all negatives (impostors and spoofing attacks)

    Returns
    -------
    far_w_errors: array_like
        FAR_w
    wer_wb_errors: array_like
        WER_wb
    """

    far_w_errors = numpy.ndarray((beta.size, omega.size), "float64")
    wer_wb_errors = numpy.ndarray((beta.size, omega.size), "float64")

    for bindex, b in enumerate(beta):
        errors = [
            error_rates_at_weight(
                licit_neg,
                licit_pos,
                spoof_neg,
                spoof_pos,
                w,
                thresholds[bindex, windex],
                b,
            )
            for windex, w in enumerate(omega)
        ]
        far_w_errors[bindex, :] = [errors[i][3] for i in range(len(errors))]
        wer_wb_errors[bindex, :] = [errors[i][4] for i in range(len(errors))]

    return far_w_errors, wer_wb_errors


def all_error_rates(
    licit_neg, licit_pos, spoof_neg, spoof_pos, thresholds, omega, beta
):
    """Calculates several error rates: FAR_w and WER_wb for the given weights
    (omega and beta) and thresholds (the thresholds need to be computed first
    using the method: epsc_thresholds() before passing to this method)

    Parameters
    ----------
    licit_neg : array_like
        array of scores for the negatives (licit scenario)
    licit_pos : array_like
        array of scores for the positives (licit scenario)
    spoof_neg : array_like
        array of scores for the negatives (spoof scenario)
    spoof_pos : array_like
        array of scores for the positives (spoof scenario)
    thresholds : array_like
        ndarray with threshold values
    omega : array_like
        array of the omega parameter balancing between impostors
        and spoofing attacks
    beta : array_like
        array of the beta parameter balancing between real accesses
        and all negatives (impostors and spoofing attacks)

    Returns
    -------
    far_w_errors: array_like
        FAR_w
    wer_wb_errors: array_like
        WER_wb
    """

    frr_errors = numpy.ndarray((beta.size, omega.size), "float64")
    far_errors = numpy.ndarray((beta.size, omega.size), "float64")
    sfar_errors = numpy.ndarray((beta.size, omega.size), "float64")
    far_w_errors = numpy.ndarray((beta.size, omega.size), "float64")
    wer_wb_errors = numpy.ndarray((beta.size, omega.size), "float64")
    hter_wb_errors = numpy.ndarray((beta.size, omega.size), "float64")

    for bindex, b in enumerate(beta):
        errors = [
            error_rates_at_weight(
                licit_neg,
                licit_pos,
                spoof_neg,
                spoof_pos,
                w,
                thresholds[bindex, windex],
                b,
            )
            for windex, w in enumerate(omega)
        ]
        frr_errors[bindex, :] = [errors[i][0] for i in range(len(errors))]
        far_errors[bindex, :] = [errors[i][1] for i in range(len(errors))]
        sfar_errors[bindex, :] = [errors[i][2] for i in range(len(errors))]
        far_w_errors[bindex, :] = [errors[i][3] for i in range(len(errors))]
        wer_wb_errors[bindex, :] = [errors[i][4] for i in range(len(errors))]
        hter_wb_errors[bindex, :] = [errors[i][5] for i in range(len(errors))]

    return (
        frr_errors,
        far_errors,
        sfar_errors,
        far_w_errors,
        wer_wb_errors,
        hter_wb_errors,
    )


def calc_aue(
    licit_neg,
    licit_pos,
    spoof_neg,
    spoof_pos,
    thresholds,
    omega,
    beta,
    l_bound=0,
    h_bound=1,
    var_param="omega",
):
    """Calculates AUE of EPSC for the given thresholds and weights

    Keyword arguments:

    - licit_neg - numpy.array of scores for the negatives (licit scenario)
    - licit_pos - numpy.array of scores for the positives (licit scenario)
    - spoof_neg - numpy.array of scores for the negatives (spoof scenario)
    - spoof_pos - numpy.array of scores for the positives (spoof scenario)
    - l_bound - lower bound of integration
    - h_bound - higher bound of integration
    - points - number of points to calculate EPSC
    - criteria - the decision threshold criteria ('eer', 'wer' or 'min-hter')
    - var_param - name of the parameter which is varied on the abscissa
    ('omega' or 'beta')
    """

    from scipy import integrate

    if var_param == "omega":
        errors = all_error_rates(
            licit_neg, licit_pos, spoof_neg, spoof_pos, thresholds, omega, beta
        )
        weights = omega  # setting the weights to the varying parameter
    else:
        errors = all_error_rates(
            licit_neg, licit_pos, spoof_neg, spoof_pos, thresholds, omega, beta
        )
        weights = beta  # setting the weights to the varying parameter

    wer_errors = errors[4].reshape(1, errors[4].size)

    l_ind = numpy.where(weights >= l_bound)[0][0]
    h_ind = numpy.where(weights <= h_bound)[0][-1]
    aue = integrate.cumtrapz(wer_errors, weights)
    aue = numpy.append(
        [0], aue
    )  # for indexing purposes, aue is cumulative integration
    aue = aue[h_ind] - aue[l_ind]

    return aue



import numpy

import bob.bio.base
import bob.io.base.testing_utils


def test_resources():
    # get list of extensions
    extensions = bob.bio.base.extensions()
    assert isinstance(extensions, list)
    assert "bob.bio.base" in extensions


def test_sampling():
    # test selection of elements
    indices = bob.bio.base.selected_indices(100, 10)
    assert indices == list(range(5, 100, 10))

    indices = bob.bio.base.selected_indices(100, 300)
    assert indices == range(100)

    indices = bob.bio.base.selected_indices(100, None)
    assert indices == range(100)

    array = numpy.arange(100)
    elements = bob.bio.base.selected_elements(array, 10)
    assert (elements - numpy.arange(5, 100, 10) == 0.0).all()

    elements = bob.bio.base.selected_elements(array, 200)
    assert (elements - numpy.arange(100) == 0.0).all()

    elements = bob.bio.base.selected_elements(array, None)
    assert (elements - numpy.arange(100) == 0.0).all()



import bob.io.base

from bob.bio.base.utils.annotations import read_annotation_file
from bob.pipelines.sample import _ReprMixin

from .legacy import File as LegacyFile


class BioFile(LegacyFile, _ReprMixin):
    """
    A simple base class that defines basic properties of File object for the use
    in verification experiments

    Attributes
    ----------
    client_id : str or int
        The id of the client this file belongs to.
        Its type depends on your implementation.
        If you use an SQL database, this should be an SQL type like Integer or
        String.
    path : object
        see :py:class:`bob.bio.base.database.legacy.File` constructor
    file_id : object
        see :py:class:`bob.bio.base.database.legacy.File` constructor
    original_directory : str or None
        The path to the original directory of the file
    original_extension : str or None
        The extension of the original files. This attribute is deprecated.
        Please try to include the extension in the ``path`` attribute
    annotation_directory : str or None
        The path to the directory of the annotations
    annotation_extension : str or None
        The extension of annotation files. Default is ``.json``
    annotation_type : str or None
        The type of the annotation file, see
        :`bob.bio.base.utils.read_annotation_file`. Default is
        ``json``.
    """

    def __init__(
        self,
        client_id,
        path,
        file_id=None,
        original_directory=None,
        original_extension=None,
        annotation_directory=None,
        annotation_extension=None,
        annotation_type=None,
        **kwargs,
    ):
        super(BioFile, self).__init__(path, file_id, **kwargs)

        # just copy the information
        self.client_id = client_id
        """The id of the client, to which this file belongs to."""
        self.original_directory = original_directory
        self.original_extension = original_extension
        self.annotation_directory = annotation_directory
        self.annotation_extension = annotation_extension or ".json"
        self.annotation_type = annotation_type or "json"

    def load(self, original_directory=None, original_extension=None):
        """Loads the data at the specified location and using the given extension.
        Override it if you need to load differently.

        Parameters
        ----------

        original_directory: :obj:`str` (optional)
            The path to the root of the dataset structure.
            If `None`, will try to use `self.original_directory`.

        original_extension: :obj:`str` (optional)
            The filename extension of every files in the dataset.
            If `None`, will try to use `self.original_extension`.

        Returns
        -------
        object
            The loaded data (normally :py:class:`numpy.ndarray`).
        """

        if original_directory is None:
            original_directory = self.original_directory
        if original_extension is None:
            original_extension = self.original_extension
        # get the path
        path = self.make_path(
            original_directory or "", original_extension or ""
        )
        return bob.io.base.load(path)

    @property
    def annotations(self):
        path = self.make_path(
            self.annotation_directory or "", self.annotation_extension or ""
        )
        return read_annotation_file(path, annotation_type=self.annotation_type)


class BioFileSet(BioFile):
    """This class defines the minimum interface of a set of database files that needs to be exported.
    Use this class, whenever the database provides several files that belong to the same probe.
    Each file set has an id, and a list of associated files, which are of
    type :py:class:`bob.bio.base.database.BioFile` of the same client.
    The file set id can be anything hashable, but needs to be unique all over the database.

    Parameters
    ----------

    file_set_id : str or int
        A unique ID that identifies the file set.

    files : [:py:class:`bob.bio.base.database.BioFile`]
        A non-empty list of BioFile objects that should be stored inside this file.
        All files of that list need to have the same client ID.
    """

    def __init__(self, file_set_id, files, path=None, **kwargs):
        # don't accept empty file lists
        assert len(files), "Cannot create an empty BioFileSet"

        # call base class constructor
        super(BioFileSet, self).__init__(
            files[0].client_id,
            "+".join(f.path for f in files) if path is None else path,
            file_set_id,
            **kwargs,
        )

        # check that all files come from the same client
        assert all(f.client_id == self.client_id for f in files)

        # The list of files contained in this set
        self.files = files

    def __lt__(self, other):
        """Defines an order between file sets by using the order of the file set ids."""
        # compare two BioFile set objects by comparing their IDs
        return self.id < other.id



import click
import pkg_resources

from clapper.click import AliasedGroup
from click_plugins import with_plugins


@with_plugins(pkg_resources.iter_entry_points("bob.bio.pipeline.cli"))
@click.group(cls=AliasedGroup)
def pipeline():
    """Available pipelines"""
    pass



"""Runs error analysis on score sets, outputs metrics and plots"""

import logging

import click
import matplotlib.pyplot as mpl
import numpy as np

from tabulate import tabulate

import bob.measure.script.figure as measure_figure

from bob.measure import (
    f_score,
    far_threshold,
    farfrr,
    frr_threshold,
    min_weighted_error_rate_threshold,
    plot,
    ppndf,
    precision_recall,
    roc_auc_score,
)
from bob.measure.utils import get_thres, remove_nan

from . import error_utils

logger = logging.getLogger(__name__)


def clean_scores(input_scores):
    """Returns a dict with each scores groups cleaned

    Parameters
    ----------
    input_scores: dict

    Returns
    -------
    clean_scores: dict
    """
    clean_scores = {}
    for key, scores in input_scores.items():
        clean_scores[key], _, _ = remove_nan(scores)
    return clean_scores


class Metrics(measure_figure.Metrics):
    """Compute metrics from score files

    Attributes
    ----------

    names: dict {str:str}
        pairs of metrics keys and corresponding row titles to display.
    """

    def __init__(
        self,
        ctx,
        scores,
        evaluation,
        func_load,
        names={
            "fta": "Licit Failure to Acquire",
            "fmr": "Licit False Match Rate",
            "fnmr": "Licit False Non Match Rate",
            "far": "Licit False Accept Rate",
            "frr": "Licit False Reject Rate",
            "hter": "Licit Half Total Error Rate",
            "iapmr": "Attack Presentation Match Rate",
        },
        **kwargs,
    ):
        super(Metrics, self).__init__(
            ctx, scores, evaluation, func_load, names, **kwargs
        )

    def _get_all_metrics(self, idx, input_scores, input_names):
        """Compute all metrics for dev and eval scores"""
        # Parse input and remove/count failed samples (NaN)
        dev_neg, dev_neg_na, dev_neg_count = remove_nan(
            input_scores[0]["licit_neg"]
        )
        dev_pos, dev_pos_na, dev_pos_count = remove_nan(
            input_scores[0]["licit_pos"]
        )
        dev_spoof, dev_spoof_na, dev_spoof_count = remove_nan(
            input_scores[0]["spoof"]
        )
        dev_fta = (dev_neg_na + dev_pos_na + dev_spoof_na) / (
            dev_neg_count + dev_pos_count + dev_spoof_count
        )
        if self._eval:
            eval_neg, eval_neg_na, eval_neg_count = remove_nan(
                input_scores[1]["licit_neg"]
            )
            eval_pos, eval_pos_na, eval_pos_count = remove_nan(
                input_scores[1]["licit_pos"]
            )
            eval_spoof, eval_spoof_na, eval_spoof_count = remove_nan(
                input_scores[1]["spoof"]
            )
            eval_fta = (eval_neg_na + eval_pos_na + eval_spoof_na) / (
                eval_neg_count + eval_pos_count + eval_spoof_count
            )
        dev_file = input_names[0]

        # Compute threshold on dev set
        threshold = (
            self.get_thres(self._criterion, dev_neg, dev_pos, self._far)
            if self._thres is None
            else self._thres[idx]
        )

        title = self._legends[idx] if self._legends is not None else None
        if self._thres is None:
            far_str = ""
            if self._criterion == "far" and self._far is not None:
                far_str = str(self._far)
            click.echo(
                f"[Min. criterion: {self._criterion.upper()} {far_str}] "
                f"Threshold on Development set `{title or dev_file}`: {threshold:e}",
                file=self.log_file,
            )
        else:
            click.echo(
                "[Min. criterion: user provided] Threshold on "
                f"Development set `{dev_file or title}`: {threshold:e}",
                file=self.log_file,
            )

        res = []
        res.append(
            self._strings(
                self._numbers(dev_neg, dev_pos, dev_spoof, threshold, dev_fta)
            )
        )

        if self._eval:
            # computes statistics for the eval set based on the threshold a
            # priori computed on the dev set
            res.append(
                self._strings(
                    self._numbers(
                        eval_neg, eval_pos, eval_spoof, threshold, eval_fta
                    )
                )
            )
        else:
            res.append(None)

        return res

    def _numbers(self, neg, pos, spoof, threshold, fta):
        """Computes each metric value"""
        # fpr and fnr
        fmr, fnmr = farfrr(neg, pos, threshold)
        hter = (fmr + fnmr) / 2.0
        far = fmr * (1 - fta)
        frr = fta + fnmr * (1 - fta)

        ni = neg.shape[0]  # number of impostors
        fm = int(round(fmr * ni))  # number of false accepts
        nc = pos.shape[0]  # number of clients
        fnm = int(round(fnmr * nc))  # number of false rejects

        # precision and recall
        precision, recall = precision_recall(neg, pos, threshold)

        # f_score
        f1_score = f_score(neg, pos, threshold, 1)

        # AUC ROC
        auc = roc_auc_score(neg, pos)
        auc_log = roc_auc_score(neg, pos, log_scale=True)

        # IAPMR at threshold
        iapmr, _ = farfrr(spoof, [0.0], threshold)
        spoof_total = len(spoof)
        spoof_match = int(round(iapmr * spoof_total))

        return {
            "fta": fta,
            "fmr": fmr,
            "fnmr": fnmr,
            "hter": hter,
            "far": far,
            "frr": frr,
            "fm": fm,
            "ni": ni,
            "fnm": fnm,
            "nc": nc,
            "precision": precision,
            "recall": recall,
            "f1_score": f1_score,
            "auc": auc,
            "auc_log": auc_log,
            "iapmr": iapmr,
            "spoof_match": spoof_match,
            "spoof_total": spoof_total,
        }

    def _strings(self, metrics):
        """Formats the metrics values into strings"""
        return {
            "fta": f"{100 * metrics['fta']:.{self._decimal}f}%",
            "fmr": f"{100 * metrics['fmr']:.{self._decimal}f}% ({metrics['fm']}/{metrics['ni']})",
            "fnmr": f"{100 * metrics['fnmr']:.{self._decimal}f}% ({metrics['fnm']}/{metrics['nc']})",
            "far": f"{100 * metrics['far']:.{self._decimal}f}%",
            "frr": f"{100 * metrics['frr']:.{self._decimal}f}%",
            "hter": f"{100 * metrics['hter']:.{self._decimal}f}%",
            "precision": f"{metrics['precision']:.{self._decimal}f}",
            "recall": f"{metrics['recall']:.{self._decimal}f}",
            "f1_score": f"{metrics['f1_score']:.{self._decimal}f}",
            "auc": f"{metrics['auc']:.{self._decimal}f}",
            "auc_log": f"{metrics['auc_log']:.{self._decimal}f}",
            "iapmr": f"{100 * metrics['iapmr']:.{self._decimal}f}% ({metrics['spoof_match']}/{metrics['spoof_total']})",
        }

    def compute(self, idx, input_scores, input_names):
        """Compute metrics thresholds and tables for given system inputs"""
        # Title and headers
        title = self._legends[idx] if self._legends is not None else None
        headers = ["" or title, "Dev. %s" % input_names[0]]
        if self._eval and input_scores[1] is not None:
            headers.append("eval % s" % input_names[1])

        # Tables rows
        all_metrics = self._get_all_metrics(idx, input_scores, input_names)
        headers = [" " or title, "Development"]

        rows = []
        for key, name in self.names.items():
            if key not in all_metrics[0]:
                logger.warning(f"{key} not present in metrics.")
            rows.append([name, all_metrics[0].get(key, "N/A")])

        if self._eval:
            # computes statistics for the eval set based on the threshold a
            # priori
            headers.append("Evaluation")
            for row, key in zip(rows, self.names.keys()):
                row.append(all_metrics[1].get(key, "N/A"))

        click.echo(tabulate(rows, headers, self._tablefmt), file=self.log_file)


def _iapmr_dot(threshold, iapmr, real_data, **kwargs):
    # plot a dot on threshold versus IAPMR line and show IAPMR as a number
    axlim = mpl.axis()
    mpl.plot(threshold, 100.0 * iapmr, "o", color="C3", **kwargs)
    if not real_data:
        mpl.annotate(
            "IAPMR at\noperating point",
            xy=(threshold, 100.0 * iapmr),
            xycoords="data",
            xytext=(0.85, 0.6),
            textcoords="axes fraction",
            color="black",
            size="large",
            arrowprops=dict(facecolor="black", shrink=0.05, width=2),
            horizontalalignment="center",
            verticalalignment="top",
        )
    else:
        mpl.text(
            threshold + (threshold - axlim[0]) / 12,
            100.0 * iapmr,
            "%.1f%%" % (100.0 * iapmr,),
            color="C3",
        )


def _iapmr_line_plot(scores, n_points=100, **kwargs):
    axlim = mpl.axis()
    step = (axlim[1] - axlim[0]) / float(n_points)
    thres = [(k * step) + axlim[0] for k in range(2, n_points - 1)]
    mix_prob_y = []
    for k in thres:
        mix_prob_y.append(100.0 * error_utils.calc_pass_rate(k, scores))

    mpl.plot(thres, mix_prob_y, label="IAPMR", color="C3", **kwargs)


def _iapmr_plot(scores, threshold, iapmr, real_data, **kwargs):
    _iapmr_dot(threshold, iapmr, real_data, **kwargs)
    _iapmr_line_plot(scores, n_points=100, **kwargs)


class HistVuln(measure_figure.Hist):
    """Histograms for vulnerability"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(HistVuln, self).__init__(
            ctx, scores, evaluation, func_load, nhist_per_system=3
        )

    def _setup_hist(self, neg, pos):
        self._title_base = " "
        self._density_hist(pos[0], n=0, label="Genuine", color="C2")
        self._density_hist(
            neg[0], n=1, label="Zero-effort impostors", alpha=0.8, color="C0"
        )
        self._density_hist(
            neg[1],
            n=2,
            label="Presentation attack",
            alpha=0.4,
            color="C7",
            hatch="\\\\",
        )

    def _get_neg_pos_thres(self, idx, input_scores, input_names):
        """Get scores and threshold for the given system at index idx for vuln

        Returns
        -------
        dev_neg, dev_pos, eval_neg, eval_pos: list of arrays
            The scores negatives and positives for each set. Each element
            contains two lists: licit [0] and spoof [1]
        threshold: int
            The value of the threshold computed on the `dev` set licit scores.
        """

        dev_scores = clean_scores(input_scores[0])
        if self._eval:
            eval_scores = clean_scores(input_scores[1])
        else:
            eval_scores = {"licit_neg": [], "licit_pos": [], "spoof": []}

        threshold = (
            get_thres(
                self._criterion,
                dev_scores["licit_neg"],
                dev_scores["licit_pos"],
            )
            if self._thres is None
            else self._thres[idx]
        )
        return (
            [
                dev_scores["licit_neg"],
                dev_scores["spoof"],
            ],
            [dev_scores["licit_pos"]],
            [
                eval_scores["licit_neg"],
                eval_scores["spoof"],
            ],
            [
                eval_scores["licit_pos"],
            ],
            threshold,
        )

    def _lines(self, threshold, label, neg, pos, idx, **kwargs):
        spoof = neg[1]
        neg = neg[0]
        pos = pos[0]
        # plot EER treshold vertical line
        super(HistVuln, self)._lines(threshold, label, neg, pos, idx, **kwargs)

        if "iapmr_line" not in self._ctx.meta or self._ctx.meta["iapmr_line"]:
            # Plot iapmr_line (accepted PA vs threshold)
            iapmr, _ = farfrr(spoof, [0.0], threshold)
            ax2 = mpl.twinx()
            # we never want grid lines on axis 2
            ax2.grid(False)
            real_data = self._ctx.meta.get("real_data", True)
            _iapmr_plot(spoof, threshold, iapmr, real_data=real_data)
            n = idx % self._step_print
            col = n % self._ncols
            rest_print = (
                self.n_systems - int(idx / self._step_print) * self._step_print
            )
            if col == self._ncols - 1 or n == rest_print - 1:
                ax2.set_ylabel("IAPMR (%)", color="C3")
            ax2.tick_params(axis="y", colors="C3")
            ax2.yaxis.label.set_color("C3")
            ax2.spines["right"].set_color("C3")


class Epc(measure_figure.PlotBase):
    """Handles the plotting of EPC"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(Epc, self).__init__(ctx, scores, evaluation, func_load)
        self._iapmr = self._ctx.meta.get("iapmr", True)
        self._titles = self._titles or [
            "EPC and IAPMR" if self._iapmr else "EPC"
        ]
        self._x_label = self._x_label or "Weight $\\beta$"
        self._y_label = self._y_label or "HTER (%)"
        self._eval = True  # always eval data with EPC
        self._split = False
        self._nb_figs = 1

        if self._min_arg != 2:
            raise click.BadParameter(
                "You must provide 2 scores files: " "scores-{dev,eval}.csv"
            )

    def compute(self, idx, input_scores, input_names):
        """Plot EPC with IAPMR for vuln"""
        dev_scores = clean_scores(input_scores[0])
        if self._eval:
            eval_scores = clean_scores(input_scores[1])
        else:
            eval_scores = {"licit_neg": [], "licit_pos": [], "spoof": []}

        mpl.gcf().clear()
        mpl.grid()
        logger.info(f"EPC using {input_names[0]} and {input_names[1]}")
        plot.epc(
            dev_scores["licit_neg"],
            dev_scores["licit_pos"],
            eval_scores["licit_neg"],
            eval_scores["licit_pos"],
            self._points,
            color="C0",
            linestyle=self._linestyles[idx],
            label=self._label("HTER (licit)", idx),
        )
        mpl.xlabel(self._x_label)
        mpl.ylabel(self._y_label)
        if self._iapmr:
            ax1 = mpl.gca()
            mpl.gca().set_axisbelow(True)
            prob_ax = mpl.gca().twinx()
            step = 1.0 / float(self._points)
            thres = [float(k * step) for k in range(self._points)]
            thres.append(1.0)
            apply_thres = [
                min_weighted_error_rate_threshold(
                    dev_scores["licit_neg"], dev_scores["licit_pos"], t
                )
                for t in thres
            ]
            mix_prob_y = []
            for k in apply_thres:
                mix_prob_y.append(
                    100.0 * error_utils.calc_pass_rate(k, eval_scores["spoof"])
                )

            logger.info(
                f"IAPMR in EPC plot using {input_names[0]}, {input_names[1]}"
            )
            mpl.plot(
                thres,
                mix_prob_y,
                label=self._label("IAPMR (spoof)", idx),
                color="C3",
            )

            prob_ax.tick_params(axis="y", colors="C3")
            prob_ax.yaxis.label.set_color("C3")
            prob_ax.spines["right"].set_color("C3")
            prob_ax.set_ylabel("IAPMR (%)", color="C3")
            prob_ax.set_axisbelow(True)
            ax1.yaxis.label.set_color("C0")
            ax1.tick_params(axis="y", colors="C0")
            ax1.spines["left"].set_color("C0")
            mpl.sca(ax1)


class Epsc(measure_figure.GridSubplot):
    """Handles the plotting of EPSC"""

    def __init__(self, ctx, scores, func_load, criteria, var_param, **kwargs):
        evaluation = ctx.meta.get("evaluation", True)
        super(Epsc, self).__init__(ctx, scores, evaluation, func_load)
        self._iapmr = self._ctx.meta.get("iapmr", False)
        self._wer = self._ctx.meta.get("wer", True)
        self._criteria = criteria or "eer"
        self._var_param = var_param or "omega"
        self._fixed_params = ctx.meta.get("fixed_params", [0.5])
        self._nb_subplots = 2 if (self._wer and self._iapmr) else 1
        if len(self._titles) < self._nb_figs * self._nb_subplots:
            self._titles = [
                v for v in self._titles for _ in range(self._nb_subplots)
            ]
        self._eval = True  # always eval data with EPSC
        self._split = False
        self._nb_figs = 1
        self._sampling = ctx.meta.get("sampling", 5)
        self._axis1 = None
        self._axis2 = None

        if self._min_arg != 2:
            raise click.BadParameter(
                "You must provide 2 scores files: " "scores-{dev,eval}.csv"
            )

        self._ncols = 1 if self._iapmr else 0
        self._ncols += 1 if self._wer else 0

    def compute(self, idx, input_scores, input_names):
        """Plot EPSC for vuln"""
        dev_scores = clean_scores(input_scores[0])
        if self._eval:
            eval_scores = clean_scores(input_scores[1])
        else:
            eval_scores = {"licit_neg": [], "licit_pos": [], "spoof": []}

        merge_sys = (
            self._fixed_params is None or len(self._fixed_params) == 1
        ) and self.n_systems > 1
        legend = ""
        if self._legends is not None and idx < len(self._legends):
            legend = self._legends[idx]
        elif self.n_systems > 1:
            legend = "Sys%d" % (idx + 1)

        if self._axis1 is None:
            # axes should only be created once
            self._axis1 = self.create_subplot(0)
            if self._ncols == 2:
                self._axis2 = self.create_subplot(1)
            else:
                self._axis2 = self._axis1
        points = 10
        for pi, fp in enumerate(self._fixed_params):
            if merge_sys:
                assert pi == 0
                pi = idx
            if self._var_param == "omega":
                omega, beta, thrs = error_utils.epsc_thresholds(
                    dev_scores["licit_neg"],
                    dev_scores["licit_pos"],
                    dev_scores["spoof"],
                    dev_scores["licit_pos"],
                    points=points,
                    criteria=self._criteria,
                    beta=fp,
                )
            else:
                omega, beta, thrs = error_utils.epsc_thresholds(
                    dev_scores["licit_neg"],
                    dev_scores["licit_pos"],
                    dev_scores["spoof"],
                    dev_scores["licit_pos"],
                    points=points,
                    criteria=self._criteria,
                    omega=fp,
                )

            errors = error_utils.all_error_rates(
                eval_scores["licit_neg"],
                eval_scores["licit_pos"],
                eval_scores["spoof"],
                eval_scores["licit_pos"],
                thrs,
                omega,
                beta,
            )  # error rates are returned in a list in the
            # following order: frr, far, IAPMR, far_w, wer_w
            mpl.sca(self._axis1)
            # Between zero-effort impostors and Presentation attacks
            if self._wer:
                logger.debug(
                    f"Plotting EPSC: WER for system {idx+1}, fix param {pi}: {fp}"
                )
                set_title = (
                    self._titles[(idx // self.n_systems) * self._nb_subplots]
                    if self._titles
                    else None
                )
                display = (
                    set_title.replace(" ", "")
                    if set_title is not None
                    else True
                )
                wer_title = set_title or "EPSC"
                if display:
                    self._axis1.set_title(wer_title)
                base = f"({legend}) " if legend.strip() else ""
                if self._var_param == "omega":
                    label = f"{base}$\\beta={fp:.1f}$"
                    self._axis1.plot(
                        omega,
                        100.0 * errors[4].flatten(),
                        color=self._colors[pi],
                        linestyle="-",
                        label=label,
                    )
                    self._axis1.set_xlabel(self._x_label or "Weight $\\omega$")
                else:
                    label = f"{base}$\\omega={fp:.1f}$"
                    self._axis1.plot(
                        beta,
                        100.0 * errors[4].flatten(),
                        color=self._colors[pi],
                        linestyle="-",
                        label=label,
                    )
                    self._axis1.set_xlabel(self._x_label or "Weight $\\beta$")
                self._axis1.set_ylabel(
                    self._y_label or "WER$_{\\omega,\\beta}$ (%)"
                )
                self._axis1.grid(True)
                self._axis1.legend(loc=self._legend_loc)

            if self._iapmr:
                logger.debug(
                    f"Plotting EPSC: IAPMR for system {idx+1}, fix param {pi}: {fp}"
                )
                mpl.sca(self._axis2)
                set_title = (
                    self._titles[
                        (idx // self.n_systems) * self._nb_subplots + 1
                    ]
                    if self._titles
                    else None
                )
                display = (
                    set_title.replace(" ", "")
                    if set_title is not None
                    else True
                )
                iapmr_title = set_title or "EPSC"
                if display:
                    self._axis2.set_title(iapmr_title)
                base = f"({legend}) " if legend.strip() else ""
                if self._var_param == "omega":
                    label = f"{base} $\\beta={fp:.1f}$"
                    self._axis2.plot(
                        omega,
                        100.0 * errors[2].flatten(),
                        color=self._colors[pi],
                        linestyle="-",
                        label=label,
                    )
                    self._axis2.set_xlabel(self._x_label or "Weight $\\omega$")
                else:
                    label = f"{base} $\\omega={fp:.1f}$"
                    self._axis2.plot(
                        beta,
                        100.0 * errors[2].flatten(),
                        linestyle="-",
                        color=self._colors[pi],
                        label=label,
                    )
                    self._axis2.set_xlabel(self._x_label or "Weight $\\beta$")

                self._axis2.set_ylabel(self._y_label or "IAPMR (%)")
                self._axis2.grid(True)
                self._axis2.legend(loc=self._legend_loc)

    def end_process(self):
        """Sets the legend."""
        if self._end_setup_plot:
            for i in range(self._nb_figs):
                fig = mpl.figure(i + 1)
                if self._disp_legend:
                    mpl.legend(loc=self._legend_loc)
                self._pdf_page.savefig(fig)
        self._end_setup_plot = False
        super().end_process()


class Epsc3D(Epsc):
    """3D EPSC plots for vuln"""

    def __init__(self, ctx, scores, func_load, criteria, var_param, **kwargs):
        super().__init__(ctx, scores, func_load, criteria, var_param, **kwargs)
        if self._nb_subplots != 1:
            raise ValueError(
                "You cannot plot more than one type of plot (WER or IAPMR)."
            )

    def compute(self, idx, input_scores, input_names):
        """Implements plots"""
        dev_scores = clean_scores(input_scores[0])
        if self._eval:
            eval_scores = clean_scores(input_scores[1])
        else:
            eval_scores = {"licit_neg": [], "licit_pos": [], "spoof": []}

        default_title = "EPSC 3D"
        title = self._titles[idx] if self._titles else default_title

        mpl.rcParams.pop("key", None)

        points = self._sampling or 5

        # Compute threshold values on dev
        omega, beta, thrs = error_utils.epsc_thresholds(
            dev_scores["licit_neg"],
            dev_scores["licit_pos"],
            dev_scores["spoof"],
            dev_scores["licit_pos"],
            points=points,
            criteria=self._criteria,
        )

        # Compute errors on eval
        errors = error_utils.all_error_rates(
            eval_scores["licit_neg"],
            eval_scores["licit_pos"],
            eval_scores["spoof"],
            eval_scores["licit_pos"],
            thrs,
            omega,
            beta,
        )
        # error rates are returned in a list as 2D numpy.ndarrays in
        # the following order: frr, far, IAPMR, far_w, wer_wb, hter_wb
        wer_errors = 100 * errors[2 if self._iapmr else 4]

        if not self._axis1:
            self._axis1 = mpl.gcf().add_subplot(111, projection="3d")

        W, B = np.meshgrid(omega, beta)

        label = self._legends[idx] if self._legends else f"Sys {idx+1}"
        self._axis1.plot_wireframe(
            W,
            B,
            wer_errors,
            color=self._colors[idx],
            antialiased=False,
            label=label,
        )

        if self._iapmr:
            self._axis1.azim = -30
            self._axis1.elev = 50

        self._axis1.set_xlabel(self._x_label or r"Weight $\omega$")
        self._axis1.set_ylabel(self._y_label or r"Weight $\beta$")
        self._axis1.set_zlabel(
            r"WER$_{\omega,\beta}$ (%)" if self._wer else "IAPMR (%)"
        )

        if title.replace(" ", ""):
            mpl.title(title)


class BaseVulnDetRoc(measure_figure.PlotBase):
    """Base for DET and ROC"""

    def __init__(self, ctx, scores, evaluation, func_load, real_data, no_spoof):
        super(BaseVulnDetRoc, self).__init__(ctx, scores, evaluation, func_load)
        self._no_spoof = no_spoof
        self._fnmrs_at = ctx.meta.get("fnmr", [])
        self._fnmrs_at = [] if self._fnmrs_at is None else self._fnmrs_at
        self._real_data = True if real_data is None else real_data
        self._min_dig = -4 if self._min_dig is None else self._min_dig
        self._tpr = ctx.meta.get("tpr", True)

    def compute(self, idx, input_scores, input_names):
        """Implements plots"""
        dev_scores = clean_scores(input_scores[0])
        if self._eval:
            eval_scores = clean_scores(input_scores[1])
        else:
            eval_scores = {"licit_neg": [], "licit_pos": [], "spoof": []}

        mpl.figure(1)
        if self._eval:
            logger.info(f"dev curve using {input_names[0]}")
            self._plot(
                dev_scores["licit_neg"],
                dev_scores["licit_pos"],
                dev_scores["spoof"],
                npoints=self._points,
                tpr=self._tpr,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev", idx),
                alpha=self._alpha,
            )
            if not self._fnmrs_at:
                logger.info("Plotting fnmr line at dev eer threshold for dev")
                dev_threshold = get_thres(
                    criter="eer",
                    neg=dev_scores["licit_neg"],
                    pos=dev_scores["licit_pos"],
                )
                _, fnmr_at_dev_threshold = farfrr(
                    [0.0], dev_scores["licit_pos"], dev_threshold
                )
            fnmrs_dev = self._fnmrs_at or [fnmr_at_dev_threshold]
            self._draw_fnmrs(idx, dev_scores, fnmrs_dev)

            if self._split:
                mpl.figure(2)

            # Add the eval plot
            linestyle = "--" if not self._split else self._linestyles[idx]
            logger.info(f"eval curve using {input_names[1]}")
            self._plot(
                eval_scores["licit_neg"],
                eval_scores["licit_pos"],
                eval_scores["spoof"],
                linestyle=linestyle,
                npoints=self._points,
                tpr=self._tpr,
                min_far=self._min_dig,
                color=self._colors[idx],
                label=self._label("eval", idx),
                alpha=self._alpha,
            )
            if not self._fnmrs_at:
                logger.info("printing fnmr at dev eer threshold for eval")
                _, fnmr_at_dev_threshold = farfrr(
                    [0.0], eval_scores["licit_pos"], dev_threshold
                )
            fnmrs_dev = self._fnmrs_at or [fnmr_at_dev_threshold]
            self._draw_fnmrs(idx, eval_scores, fnmrs_dev, True)

        # Only dev scores available
        else:
            logger.info(f"dev curve using {input_names[0]}")
            self._plot(
                dev_scores["licit_neg"],
                dev_scores["licit_pos"],
                dev_scores["spoof"],
                npoints=self._points,
                tpr=self._tpr,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev", idx),
                alpha=self._alpha,
            )
            if not self._fnmrs_at:
                logger.info("Plotting fnmr line at dev eer threshold for dev")
                dev_threshold = get_thres(
                    criter="eer",
                    neg=dev_scores["licit_neg"],
                    pos=dev_scores["licit_pos"],
                )
                _, fnmr_at_dev_threshold = farfrr(
                    [0.0], dev_scores["licit_pos"], dev_threshold
                )
            fnmrs_dev = self._fnmrs_at or [fnmr_at_dev_threshold]
            self._draw_fnmrs(idx, dev_scores, fnmrs_dev)

    def _get_farfrr(self, x, y, thres):
        return None, None

    def _plot(self, x, y, s, npoints, **kwargs):
        pass

    def _draw_fnmrs(self, idx, scores, fnmrs=[], eval=False):
        pass


class DetVuln(BaseVulnDetRoc):
    """DET for vuln"""

    def __init__(self, ctx, scores, evaluation, func_load, real_data, no_spoof):
        super(DetVuln, self).__init__(
            ctx, scores, evaluation, func_load, real_data, no_spoof
        )
        self._x_label = self._x_label or "FMR (%)"
        self._y_label = self._y_label or "FNMR (%)"
        self._semilogx = ctx.meta.get("semilogx", False)
        add = ""
        if not self._titles:
            self._titles = [""] * self._nb_figs
        if not self._no_spoof:
            add = " and overlaid SPOOF scenario"
        for i, t in enumerate(self._titles):
            if self._eval and (i % 2):
                dev_eval = ", eval group"
            elif self._eval:
                dev_eval = ", dev group"
            else:
                dev_eval = ""
            self._titles[i] = t or ("DET: LICIT" + add + dev_eval)
        self._legend_loc = self._legend_loc or "upper right"

    def _set_axis(self):
        if self._axlim is not None and None not in self._axlim:
            plot.det_axis(self._axlim)
        else:
            plot.det_axis([0.01, 99, 0.01, 99])

    def _get_farfrr(self, x, y, thres):
        points = farfrr(x, y, thres)
        return points, [ppndf(i) for i in points]

    def _plot(self, x, y, s, npoints, **kwargs):
        logger.info("Plotting DET")
        plot.det(
            x,
            y,
            npoints,
            min_far=self._min_dig,
            color=kwargs.get("color"),
            linestyle=kwargs.get("linestyle"),
            label=kwargs.get("label"),
        )
        if not self._no_spoof and s is not None:
            ax1 = mpl.gca()
            ax2 = ax1.twiny()
            ax2.set_xlabel("IAPMR (%)", color="C3")
            ax2.tick_params(
                axis="x",
                colors="C3",
                labelrotation=self._x_rotation,
                labelcolor="C3",
            )
            # Prevent tick labels overlap
            ax2.tick_params(axis="both", which="major", labelsize="x-small")
            ax1.tick_params(axis="both", which="major", labelsize="x-small")
            ax2.spines["top"].set_color("C3")
            plot.det(
                s,
                y,
                npoints,
                min_far=self._min_dig,
                color="C3",
                linestyle=":",
                label="Spoof " + kwargs.get("label"),
            )
            self._set_axis()
            mpl.sca(ax1)

    def _draw_fnmrs(self, idx, scores, fnmrs=[], eval=False):
        for line in fnmrs:
            thres_baseline = frr_threshold(
                scores["licit_neg"], scores["licit_pos"], line
            )

            axlim = mpl.axis()

            farfrr_licit, farfrr_licit_det = self._get_farfrr(
                scores["licit_neg"], scores["licit_pos"], thres_baseline
            )
            if farfrr_licit is None:
                return

            farfrr_spoof, farfrr_spoof_det = self._get_farfrr(
                scores["spoof"],
                scores["licit_pos"],
                frr_threshold(
                    scores["spoof"], scores["licit_pos"], farfrr_licit[1]
                ),
            )

            if not self._real_data:
                # Takes specified FNMR value as EER
                mpl.axhline(
                    y=farfrr_licit_det[1],
                    xmin=axlim[2],
                    xmax=axlim[3],
                    color="k",
                    linestyle="--",
                    label="%s @ EER" % self._y_label,
                )
            else:
                mpl.axhline(
                    y=farfrr_licit_det[1],
                    xmin=axlim[0],
                    xmax=axlim[1],
                    color="k",
                    linestyle="--",
                    label="%s = %.2f%%" % ("FNMR", farfrr_licit[1] * 100),
                )

            if not self._real_data:
                label_licit = "%s @ operating point" % self._x_label
                label_spoof = "IAPMR @ operating point"
            else:
                label_licit = "FMR=%.2f%%" % (farfrr_licit[0] * 100)
                label_spoof = "IAPMR=%.2f%%" % (farfrr_spoof[0] * 100)

            # Annotations and drawing of the points
            text_x_offset = 2
            text_y_offset = 5
            # Licit
            mpl.annotate(
                xy=(farfrr_licit_det[0], farfrr_licit_det[1]),
                text=label_licit,
                xytext=(text_x_offset, text_y_offset),
                textcoords="offset points",
                fontsize="small",
            )
            mpl.plot(
                farfrr_licit_det[0],
                farfrr_licit_det[1],
                "o",
                color=self._colors[idx],
            )  # FAR point, licit scenario
            # Spoof
            axlim = self._axlim or [0.01, 99, 0.1, 99]
            if (
                farfrr_spoof_det[0] > axlim[0]
                and farfrr_spoof_det[0] < axlim[1]
            ):
                mpl.annotate(
                    xy=(farfrr_spoof_det[0], farfrr_spoof_det[1]),
                    text=label_spoof,
                    xytext=(text_x_offset, text_y_offset),
                    textcoords="offset points",
                    fontsize="small",
                )
                mpl.plot(
                    farfrr_spoof_det[0],
                    farfrr_spoof_det[1],
                    "o",
                    color="C3",
                )  # FAR point, spoof scenario
            else:
                logger.warning(
                    f"The IAPMR for an FNMR of {line} is outside the plot."
                )


class RocVuln(BaseVulnDetRoc):
    """ROC for vuln"""

    def __init__(self, ctx, scores, evaluation, func_load, real_data, no_spoof):
        super(RocVuln, self).__init__(
            ctx, scores, evaluation, func_load, real_data, no_spoof
        )
        self._x_label = self._x_label or "FMR"
        self._y_label = self._y_label or "1 - FNMR"
        self._semilogx = ctx.meta.get("semilogx", True)
        add = ""
        if not self._titles:
            self._titles = [""] * self._nb_figs
        if not self._no_spoof:
            add = " and overlaid SPOOF scenario"
        for i, t in enumerate(self._titles):
            if self._eval and (i % 2):
                dev_eval = ", eval group"
            elif self._eval:
                dev_eval = ", dev group"
            else:
                dev_eval = ""
            self._titles[i] = t or ("ROC: LICIT" + add + dev_eval)
        if self._legend_loc == "best":
            self._legend_loc = (
                "lower right" if self._semilogx else "upper right"
            )

    def _plot(self, x, y, s, npoints, **kwargs):
        logger.info("Plotting ROC")
        plot.roc(
            x,
            y,
            npoints=npoints,
            semilogx=self._semilogx,
            tpr=self._tpr,
            min_far=self._min_dig,
            color=kwargs.get("color"),
            linestyle=kwargs.get("linestyle"),
            label=kwargs.get("label"),
        )
        if not self._no_spoof and s is not None:
            ax1 = mpl.gca()
            ax1.plot(
                [0],
                [0],
                linestyle=":",
                color="C3",
                label="Spoof " + kwargs.get("label"),
            )
            ax2 = ax1.twiny()
            ax2.set_xlabel("IAPMR (%)", color="C3")
            mpl.xticks(rotation=self._x_rotation)
            ax2.tick_params(
                axis="x",
                colors="C3",
                labelrotation=self._x_rotation,
                labelcolor="C3",
            )
            ax2.spines["top"].set_color("C3")
            plot.roc(
                s,
                y,
                npoints=npoints,
                semilogx=self._semilogx,
                tpr=self._tpr,
                min_far=self._min_dig,
                color="C3",
                linestyle=":",
                label="Spoof " + kwargs.get("label"),
            )
            self._set_axis()
            mpl.sca(ax1)

    def _get_farfrr(self, x, y, thres):
        points = farfrr(x, y, thres)
        points2 = (points[0], 1 - points[1])
        return points, points2

    def _draw_fnmrs(self, idx, scores, fnmrs=[], evaluation=False):
        for line in fnmrs:
            thres_baseline = frr_threshold(
                scores["licit_neg"], scores["licit_pos"], line
            )

            axlim = mpl.axis()

            farfrr_licit, farfrr_licit_roc = self._get_farfrr(
                scores["licit_neg"], scores["licit_pos"], thres_baseline
            )
            if farfrr_licit is None:
                return

            farfrr_spoof, farfrr_spoof_roc = self._get_farfrr(
                scores["spoof"],
                scores["licit_pos"],
                frr_threshold(
                    scores["spoof"], scores["licit_pos"], farfrr_licit[1]
                ),
            )

            if not self._real_data and not evaluation:
                mpl.axhline(
                    y=farfrr_licit_roc[1],
                    xmin=axlim[2],
                    xmax=axlim[3],
                    color="k",
                    linestyle="--",
                    label=f"{self._y_label} @ EER",
                )
            elif not evaluation:
                mpl.axhline(
                    y=farfrr_licit_roc[1],
                    xmin=axlim[0],
                    xmax=axlim[1],
                    color="k",
                    linestyle="--",
                    label=f"FNMR = {farfrr_licit[1] * 100:.2f}%",
                )

            if not self._real_data:
                label_licit = f"{self._x_label} @ operating point"
                label_spoof = "IAPMR @ operating point"
            else:
                label_licit = f"FMR={farfrr_licit[0] * 100:.2f}%"
                label_spoof = f"IAPMR={farfrr_spoof[0] * 100:.2f}%"

            mpl.plot(
                farfrr_licit_roc[0],
                farfrr_licit_roc[1],
                "o",
                color=self._colors[idx],
                label=label_licit,
            )  # FAR point, licit scenario
            mpl.plot(
                farfrr_spoof_roc[0],
                farfrr_spoof_roc[1],
                "o",
                color="C3",
                label=label_spoof,
            )  # FAR point, spoof scenario


class FmrIapmr(measure_figure.PlotBase):
    """FMR vs IAPMR"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(FmrIapmr, self).__init__(ctx, scores, evaluation, func_load)
        self._eval = True  # Always ask for eval data
        self._split = False
        self._nb_figs = 1
        self._semilogx = ctx.meta.get("semilogx", False)
        if not self._titles:
            self._titles = [""] * self._nb_figs
        for i, t in enumerate(self._titles):
            self._titles[i] = t or "FMR vs IAPMR"
        self._x_label = self._x_label or "FMR"
        self._y_label = self._y_label or "IAPMR"
        if self._min_arg != 2:
            raise click.BadParameter(
                "You must provide 2 scores files: " "scores-{dev,eval}.csv"
            )

    def compute(self, idx, input_scores, input_names):
        """Implements plots"""
        dev_scores = clean_scores(input_scores[0])
        if self._eval:
            eval_scores = clean_scores(input_scores[1])
        fmr_list = np.linspace(0, 1, 100)
        iapmr_list = []
        for i, fmr in enumerate(fmr_list):
            thr = far_threshold(
                dev_scores["licit_neg"], dev_scores["licit_pos"], fmr
            )
            iapmr_list.append(farfrr(eval_scores["spoof"], [0.0], thr)[0])
            # re-calculate fmr since threshold might give a different result
            # for fmr.
            fmr_list[i], _ = farfrr(eval_scores["licit_neg"], [0.0], thr)
        label = (
            self._legends[idx]
            if self._legends is not None
            else f"system {idx+1}"
        )
        logger.info(f"Plot FmrIapmr using: {input_names[1]}")
        if self._semilogx:
            mpl.semilogx(fmr_list, iapmr_list, label=label)
        else:
            mpl.plot(fmr_list, iapmr_list, label=label)



# isort: skip_file
from .utils import *  # noqa: F401,F403
from . import database  # noqa: F401
from . import preprocessor  # noqa: F401
from . import extractor  # noqa: F401
from . import algorithm  # noqa: F401
from . import annotator  # noqa: F401
from . import pipelines  # noqa: F401

from . import script  # noqa: F401
from . import score  # noqa: F401

# gets sphinx autodoc done right - don't remove it
__all__ = [_ for _ in dir() if not _.startswith("_")]



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""A few checks at the Verification Filelist database.
"""

import os

from typing import Any

import numpy as np
import pytest

from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer

import bob.io.base
import bob.io.base.testing_utils

from bob.bio.base.algorithm import Distance
from bob.bio.base.config.dummy.database import database as ATNT_DATABASE
from bob.bio.base.database import (
    AnnotationsLoader,
    CSVDatabase,
    FileListBioDatabase,
    FileSampleLoader,
)
from bob.bio.base.pipelines.pipelines import PipelineSimple
from bob.pipelines import Sample, SampleSet, wrap

legacy_example_dir = os.path.realpath(
    bob.io.base.testing_utils.datafile(".", __name__, "data/")
)

legacy2_example_dir = os.path.realpath(
    bob.io.base.testing_utils.datafile(".", __name__, "data/")
)


example_protocol_path = os.path.realpath(
    bob.io.base.testing_utils.datafile(
        ".", __name__, "data/example_csv_filelist"
    )
)
example_protocol_archive = os.path.realpath(
    bob.io.base.testing_utils.datafile(
        ".", __name__, "data/example_csv_filelist.tar.gz"
    )
)
atnt_protocol_path = os.path.realpath(
    bob.io.base.testing_utils.datafile(".", __name__, "data/atnt")
)


def all_sample(objects: list[Any]) -> bool:
    """Checks that all elements of ``objects`` are valid :class:`Sample` instances."""
    return all(isinstance(o, Sample) for o in objects)


def all_sample_set(objects: list[Any]) -> bool:
    """Checks that all elements of ``objects`` are :class:`SampleSet` instances.

    Also checks that all samples in these samplesets are :class:`Sample` instances.
    """

    return all(isinstance(o, SampleSet) and all_sample(o) for o in objects)


def test_csv_file_list_dev_no_metadata():
    dataset = CSVDatabase(
        name="example_csv_filelist",
        dataset_protocols_path=example_protocol_path,
        protocol="protocol_only_dev",
    )
    assert len(dataset.background_model_samples()) == 8
    assert all_sample(dataset.background_model_samples())

    assert len(dataset.references()) == 2
    assert all_sample_set(dataset.references())
    assert all(len(r) == 4 for r in dataset.references())

    assert len(dataset.probes()) == 2
    assert all_sample_set(dataset.probes())
    assert len(dataset.probes()[0]) == 4
    assert len(dataset.probes()[1]) == 6


def test_csv_file_list_dev_metadata():
    dataset = CSVDatabase(
        name="example_csv_filelist",
        dataset_protocols_path=example_protocol_path,
        protocol="protocol_only_dev_metadata",
    )
    assert len(dataset.background_model_samples()) == 8

    assert all_sample(dataset.background_model_samples())
    assert all(
        hasattr(s, "metadata_1") for s in dataset.background_model_samples()
    )
    assert all(
        hasattr(s, "metadata_2") for s in dataset.background_model_samples()
    )

    assert len(dataset.references()) == 2
    assert all_sample_set(dataset.references())
    assert all(len(r) == 4 for r in dataset.references())
    flat_references = (s for ss in dataset.references() for s in ss)
    assert all(hasattr(s, "metadata_1") for s in flat_references)
    assert all(hasattr(s, "metadata_2") for s in flat_references)

    assert len(dataset.probes()) == 2
    assert all_sample_set(dataset.probes())
    assert all(hasattr(s, "references") for s in dataset.probes())
    assert len(dataset.probes()[0]) == 4
    assert len(dataset.probes()[1]) == 6
    flat_probes = (s for ss in dataset.probes() for s in ss)
    assert all(hasattr(s, "metadata_1") for s in flat_probes)
    assert all(hasattr(s, "metadata_2") for s in flat_probes)


def test_csv_file_list_dev_eval_all_vs_all():
    annotation_directory = os.path.realpath(
        bob.io.base.testing_utils.datafile(
            ".", __name__, "data/example_csv_filelist/annotations"
        )
    )

    def run(filename):
        dataset = CSVDatabase(
            name="example_csv_filelist",
            dataset_protocols_path=filename,
            protocol="protocol_dev_eval",
            transformer=make_pipeline(
                FileSampleLoader(
                    data_loader=bob.io.base.load,
                    dataset_original_directory="non-existent",
                    extension=".nothing",
                ),
                AnnotationsLoader(
                    annotation_directory=annotation_directory,
                    annotation_extension=".json",
                    annotation_type="json",
                ),
            ),
        )

        assert dataset.score_all_vs_all
        assert len(dataset.background_model_samples()) == 8
        assert all_sample(dataset.background_model_samples())

        assert len(dataset.references()) == 2
        assert all_sample_set(dataset.references())
        assert len(dataset.references()[0]) == 4
        assert len(dataset.references()[1]) == 4

        assert len(dataset.probes()) == 2
        assert all_sample_set(dataset.probes())
        assert len(dataset.probes()[0]) == 4
        assert len(dataset.probes()[1]) == 4

        eval_references = dataset.references(group="eval")
        assert len(eval_references) == 6
        assert all_sample_set(eval_references)
        assert all(
            len(eval_references[i]) == n
            for i, n in enumerate((2, 2, 3, 1, 1, 1))
        )

        eval_probes = dataset.probes(group="eval")
        assert len(eval_probes) == 8
        assert all_sample_set(eval_probes)
        assert all(
            len(eval_probes[i]) == n
            for i, n in enumerate((2, 2, 3, 1, 1, 2, 1, 1))
        )

        assert len(dataset.all_samples(groups=None)) == 63
        assert all_sample(dataset.all_samples(groups=None))

        # Check the annotations
        for s in dataset.all_samples(groups=None):
            assert isinstance(s.annotations, dict)

        assert len(dataset.template_ids(group="dev")) == 2
        assert len(dataset.template_ids(group="eval")) == 6

        assert len(dataset.groups()) == 3

    run(example_protocol_path)
    run(example_protocol_archive)


def test_csv_file_list_dev_eval_score_norm():
    annotation_directory = os.path.realpath(
        bob.io.base.testing_utils.datafile(
            ".", __name__, "data/example_csv_filelist/annotations"
        )
    )

    def run(filename):
        znorm_dataset = CSVDatabase(
            name="example_csv_filelist",
            dataset_protocols_path=filename,
            protocol="protocol_dev_eval",
            transformer=make_pipeline(
                FileSampleLoader(
                    data_loader=bob.io.base.load,
                    dataset_original_directory="non-existent",
                    extension=".nothing",
                ),
                AnnotationsLoader(
                    annotation_directory=annotation_directory,
                    annotation_extension=".json",
                    annotation_type="json",
                ),
            ),
        )

        assert len(znorm_dataset.background_model_samples()) == 8
        assert all_sample(znorm_dataset.background_model_samples())

        assert len(znorm_dataset.references()) == 2
        assert sum(len(s) for s in znorm_dataset.references()) == 8
        assert all_sample_set(znorm_dataset.references())

        assert len(znorm_dataset.probes()) == 2
        assert sum(len(s) for s in znorm_dataset.probes()) == 8
        assert all_sample_set(znorm_dataset.references())

        assert len(znorm_dataset.references(group="eval")) == 6
        assert sum(len(s) for s in znorm_dataset.references(group="eval")) == 10
        assert all_sample_set(znorm_dataset.references(group="eval"))

        assert len(znorm_dataset.probes(group="eval")) == 8
        assert sum(len(s) for s in znorm_dataset.probes(group="eval")) == 13
        assert all_sample_set(znorm_dataset.probes(group="eval"))

        assert len(znorm_dataset.all_samples(groups=None)) == 63
        assert all_sample(znorm_dataset.all_samples(groups=None))

        # Check the annotations
        for s in znorm_dataset.all_samples(groups=None):
            assert isinstance(s.annotations, dict)

        assert len(znorm_dataset.template_ids(group="dev")) == 2
        assert len(znorm_dataset.template_ids(group="eval")) == 6
        assert len(znorm_dataset.groups()) == 3
        assert set(znorm_dataset.groups()) == set(("train", "dev", "eval"))

        # Checking ZT-Norm stuff
        assert len(znorm_dataset.treferences()) == 2
        assert len(znorm_dataset.zprobes()) == 2
        assert sum(len(s) for s in znorm_dataset.zprobes()) == 8

        assert len(znorm_dataset.treferences(proportion=0.5)) == 1
        assert len(znorm_dataset.zprobes(proportion=0.5)) == 1
        assert sum(len(s) for s in znorm_dataset.zprobes(proportion=0.5)) == 4

    run(example_protocol_path)
    run(example_protocol_archive)


def test_csv_file_list_dev_eval_sparse():
    annotation_directory = os.path.realpath(
        bob.io.base.testing_utils.datafile(
            ".", __name__, "data/example_csv_filelist/annotations"
        )
    )

    dataset = CSVDatabase(
        name="example_csv_filelist",
        dataset_protocols_path=example_protocol_path,
        protocol="protocol_dev_eval_sparse",
        transformer=make_pipeline(
            FileSampleLoader(
                data_loader=bob.io.base.load,
                dataset_original_directory="non-existent",
                extension=".nothing",
            ),
            AnnotationsLoader(
                annotation_directory=annotation_directory,
                annotation_extension=".json",
                annotation_type="json",
            ),
        ),
    )

    assert not dataset.score_all_vs_all
    assert len(dataset.background_model_samples()) == 8
    assert all_sample(dataset.background_model_samples())

    assert len(dataset.references()) == 2
    assert all_sample_set(dataset.references())

    probes = dataset.probes()
    assert len(probes) == 2
    assert sum(len(s) for s in probes) == 8

    # Here, 1 comparison per probe
    for p in probes:
        assert len(p.references) == 1
    assert all_sample_set(dataset.references())

    assert len(dataset.references(group="eval")) == 6
    assert all_sample_set(dataset.references(group="eval"))

    probes = dataset.probes(group="eval")
    assert len(probes) == 8
    assert sum(len(s) for s in probes) == 14
    assert all_sample_set(probes)
    # Here, 1 comparison per probe, EXPECT THE FIRST ONE
    for i, p in enumerate(probes):
        if i == 0:
            assert len(p.references) == 2
        else:
            assert len(p.references) == 1

    assert len(dataset.all_samples(groups=None)) == 48
    assert all_sample(dataset.all_samples(groups=None))

    # Check the annotations
    for s in dataset.all_samples(groups=None):
        assert isinstance(s.annotations, dict)

    assert len(dataset.template_ids(group="dev")) == 2
    assert len(dataset.template_ids(group="eval")) == 6

    assert len(dataset.groups()) == 3


def test_csv_file_list_atnt():
    dataset = CSVDatabase(
        name="atnt",
        dataset_protocols_path=atnt_protocol_path,
        protocol="idiap_protocol",
    )
    assert len(dataset.background_model_samples()) == 200
    assert len(dataset.references()) == 20
    assert sum(len(s) for s in dataset.references()) == 100
    assert all(len(s) == 5 for s in dataset.references())
    assert len(dataset.probes()) == 100
    assert sum(len(s) for s in dataset.probes()) == 100
    assert all(len(s) == 1 for s in dataset.probes())
    assert len(dataset.all_samples(groups=["train"])) == 200
    assert len(dataset.all_samples(groups=["dev"])) == 200
    assert len(dataset.all_samples(groups=None)) == 400


def data_loader(path):
    return bob.io.base.load(path)


def run_experiment(dataset):
    def linearize(X):
        X = np.asarray(X)
        return np.reshape(X, (X.shape[0], -1))

    # Testing it in a real recognition systems
    transformer = wrap(
        ["sample"], make_pipeline(FunctionTransformer(linearize))
    )

    pipeline_simple = PipelineSimple(transformer, Distance())

    return pipeline_simple(
        dataset.background_model_samples(),
        dataset.references(),
        dataset.probes(),
    )


def test_atnt_experiment():
    dataset = CSVDatabase(
        name="atnt",
        dataset_protocols_path=atnt_protocol_path,
        protocol="idiap_protocol",
        transformer=FileSampleLoader(
            data_loader=data_loader,
            dataset_original_directory=ATNT_DATABASE.original_directory,
            extension=".pgm",
        ),
    )

    scores = run_experiment(dataset)
    assert len(scores) == 100
    assert all(len(s) == 20 for s in scores)


####
# Testing the Legacy file list
####


def test_query():
    db = FileListBioDatabase(
        os.path.join(legacy_example_dir, "example_filelist"),
        "test",
        use_dense_probe_file_list=False,
    )

    assert (
        len(db.groups()) == 5
    )  # 5 groups (dev, eval, world, optional_world_1, optional_world_2)

    assert len(db.client_ids()) == 6  # 6 client ids for world, dev and eval
    assert len(db.client_ids(groups="world")) == 2  # 2 client ids for world
    assert (
        len(db.client_ids(groups="optional_world_1")) == 2
    )  # 2 client ids for optional world 1
    assert (
        len(db.client_ids(groups="optional_world_2")) == 2
    )  # 2 client ids for optional world 2
    assert len(db.client_ids(groups="dev")) == 2  # 2 client ids for dev
    assert len(db.client_ids(groups="eval")) == 2  # 2 client ids for eval

    assert (
        len(db.tclient_ids()) == 2
    )  # 2 client ids for T-Norm score normalization
    assert (
        len(db.zclient_ids()) == 2
    )  # 2 client ids for Z-Norm score normalization

    assert (
        len(db.model_ids_with_protocol()) == 6
    )  # 6 model ids for world, dev and eval
    assert (
        len(db.model_ids_with_protocol(groups="world")) == 2
    )  # 2 model ids for world
    assert (
        len(db.model_ids_with_protocol(groups="optional_world_1")) == 2
    )  # 2 model ids for optional world 1
    assert (
        len(db.model_ids_with_protocol(groups="optional_world_2")) == 2
    )  # 2 model ids for optional world 2
    assert (
        len(db.model_ids_with_protocol(groups="dev")) == 2
    )  # 2 model ids for dev
    assert (
        len(db.model_ids_with_protocol(groups="eval")) == 2
    )  # 2 model ids for eval

    assert (
        len(db.tmodel_ids_with_protocol()) == 2
    )  # 2 model ids for T-Norm score normalization

    assert len(db.objects(groups="world")) == 8  # 8 samples in the world set

    assert (
        len(db.objects(groups="dev", purposes="enroll")) == 8
    )  # 8 samples for enrollment in the dev set
    assert (
        len(db.objects(groups="dev", purposes="enroll", model_ids="3")) == 4
    )  # 4 samples for to enroll model '3' in the dev set
    assert (
        len(db.objects(groups="dev", purposes="enroll", model_ids="7")) == 0
    )  # 0 samples for enrolling model '7' (it is a T-Norm model)
    assert (
        len(db.objects(groups="dev", purposes="probe")) == 8
    )  # 8 samples as probes in the dev set
    assert (
        len(db.objects(groups="dev", purposes="probe", classes="client")) == 8
    )  # 8 samples as client probes in the dev set
    assert (
        len(db.objects(groups="dev", purposes="probe", classes="impostor")) == 4
    )  # 4 samples as impostor probes in the dev set

    assert (
        len(db.tobjects(groups="dev")) == 8
    )  # 8 samples for enrolling T-norm models
    assert (
        len(db.tobjects(groups="dev", model_ids="7")) == 4
    )  # 4 samples for enrolling T-norm model '7'
    assert (
        len(db.tobjects(groups="dev", model_ids="3")) == 0
    )  # 0 samples for enrolling T-norm model '3' (no T-Norm model)
    assert (
        len(db.zobjects(groups="dev")) == 8
    )  # 8 samples for Z-norm impostor accesses

    assert db.client_id_from_model_id("1", group=None) == "1"
    assert db.client_id_from_model_id("3", group=None) == "3"
    assert db.client_id_from_model_id("6", group=None) == "6"
    assert db.client_id_from_t_model_id("7", group=None) == "7"


def test_query_protocol():
    db = FileListBioDatabase(
        legacy_example_dir,
        "test",
        protocol="example_filelist",
        use_dense_probe_file_list=False,
    )

    assert (
        len(db.groups()) == 5
    )  # 5 groups (dev, eval, world, optional_world_1, optional_world_2)

    assert len(db.client_ids()) == 6  # 6 client ids for world, dev and eval
    assert (
        len(
            db.client_ids(
                groups="world",
            )
        )
        == 2
    )  # 2 client ids for world
    assert (
        len(
            db.client_ids(
                groups="optional_world_1",
            )
        )
        == 2
    )  # 2 client ids for optional world 1
    assert (
        len(
            db.client_ids(
                groups="optional_world_2",
            )
        )
        == 2
    )  # 2 client ids for optional world 2
    assert (
        len(
            db.client_ids(
                groups="dev",
            )
        )
        == 2
    )  # 2 client ids for dev
    assert (
        len(
            db.client_ids(
                groups="eval",
            )
        )
        == 2
    )  # 2 client ids for eval

    assert (
        len(db.tclient_ids()) == 2
    )  # 2 client ids for T-Norm score normalization
    assert (
        len(db.zclient_ids()) == 2
    )  # 2 client ids for Z-Norm score normalization

    assert (
        len(db.model_ids_with_protocol()) == 6
    )  # 6 model ids for world, dev and eval
    assert (
        len(
            db.model_ids_with_protocol(
                groups="world",
            )
        )
        == 2
    )  # 2 model ids for world
    assert (
        len(
            db.model_ids_with_protocol(
                groups="optional_world_1",
            )
        )
        == 2
    )  # 2 model ids for optional world 1
    assert (
        len(
            db.model_ids_with_protocol(
                groups="optional_world_2",
            )
        )
        == 2
    )  # 2 model ids for optional world 2
    assert (
        len(
            db.model_ids_with_protocol(
                groups="dev",
            )
        )
        == 2
    )  # 2 model ids for dev
    assert (
        len(
            db.model_ids_with_protocol(
                groups="eval",
            )
        )
        == 2
    )  # 2 model ids for eval

    assert (
        len(db.tmodel_ids_with_protocol()) == 2
    )  # 2 model ids for T-Norm score normalization

    assert (
        len(
            db.objects(
                groups="world",
            )
        )
        == 8
    )  # 8 samples in the world set

    assert (
        len(
            db.objects(
                groups="dev",
                purposes="enroll",
            )
        )
        == 8
    )  # 8 samples for enrollment in the dev set
    assert (
        len(
            db.objects(
                groups="dev",
                purposes="enroll",
                model_ids="3",
            )
        )
        == 4
    )  # 4 samples for to enroll model '3' in the dev set
    assert (
        len(
            db.objects(
                groups="dev",
                purposes="enroll",
                model_ids="7",
            )
        )
        == 0
    )  # 0 samples for enrolling model '7' (it is a T-Norm model)
    assert (
        len(
            db.objects(
                groups="dev",
                purposes="probe",
            )
        )
        == 8
    )  # 8 samples as probes in the dev set
    assert (
        len(
            db.objects(
                groups="dev",
                purposes="probe",
                classes="client",
            )
        )
        == 8
    )  # 8 samples as client probes in the dev set
    assert (
        len(
            db.objects(
                groups="dev",
                purposes="probe",
                classes="impostor",
            )
        )
        == 4
    )  # 4 samples as impostor probes in the dev set

    assert (
        len(
            db.tobjects(
                groups="dev",
            )
        )
        == 8
    )  # 8 samples for enrolling T-norm models
    assert (
        len(
            db.tobjects(
                groups="dev",
                model_ids="7",
            )
        )
        == 4
    )  # 4 samples for enrolling T-norm model '7'
    assert (
        len(
            db.tobjects(
                groups="dev",
                model_ids="3",
            )
        )
        == 0
    )  # 0 samples for enrolling T-norm model '3' (no T-Norm model)
    assert (
        len(db.zobjects(groups="dev")) == 8
    )  # 8 samples for Z-norm impostor accesses

    assert db.client_id_from_model_id("1", group=None) == "1"
    assert db.client_id_from_model_id("3", group=None) == "3"
    assert db.client_id_from_model_id("6", group=None) == "6"
    assert db.client_id_from_t_model_id("7", group=None) == "7"

    # check other protocols
    assert len(db.objects(protocol="non-existent")) == 0

    prot = "example_filelist2"
    assert (
        len(db.model_ids_with_protocol(protocol=prot)) == 3
    )  # 3 model ids for dev only
    with pytest.raises(ValueError):
        # eval does not exist for this protocol
        db.model_ids_with_protocol(protocol=prot, groups="eval")
    assert len(db.objects(protocol=prot, groups="dev", purposes="enroll")) == 12
    assert len(db.objects(protocol=prot, groups="dev", purposes="probe")) == 9


def test_noztnorm():
    db = FileListBioDatabase(
        os.path.join(legacy_example_dir, "example_filelist2"), "test"
    )
    assert len(db.all_files())


def test_query_dense():
    db = FileListBioDatabase(
        os.path.join(legacy_example_dir, "example_filelist"),
        "test",
        use_dense_probe_file_list=True,
    )

    assert len(db.objects(groups="world")) == 8  # 8 samples in the world set

    assert (
        len(db.objects(groups="dev", purposes="enroll")) == 8
    )  # 8 samples for enrollment in the dev set
    assert (
        len(db.objects(groups="dev", purposes="probe")) == 8
    )  # 8 samples as probes in the dev set


def test_annotation():
    db = FileListBioDatabase(
        os.path.join(legacy_example_dir, "example_filelist"),
        "test",
        use_dense_probe_file_list=False,
        annotation_directory=os.path.join(
            legacy_example_dir, "example_filelist"
        ),
        annotation_type="json",
    )
    f = [o for o in db.objects() if o.path == "data/model4_session1_sample2"][0]

    annots = db.annotations(f)

    assert annots is not None
    assert "reye" in annots
    assert "leye" in annots
    assert annots["reye"] == [20, 10]
    assert annots["leye"] == [40, 30]


def test_multiple_extensions():
    # check that the old behavior still works
    db = FileListBioDatabase(
        legacy_example_dir,
        "test",
        protocol="example_filelist",
        use_dense_probe_file_list=False,
        original_directory=os.path.join(legacy_example_dir, "example_filelist"),
        original_extension=".json",
    )

    file = bob.bio.base.database.BioFile(
        4, "data/model4_session1_sample2", "data/model4_session1_sample2"
    )

    file_name = db.original_file_name(file, True)
    assert file_name == os.path.join(
        legacy_example_dir, "example_filelist", file.path + ".json"
    )

    # check that the new behavior works as well
    db = FileListBioDatabase(
        legacy_example_dir,
        "test",
        protocol="example_filelist",
        use_dense_probe_file_list=False,
        original_directory=os.path.join(legacy_example_dir, "example_filelist"),
        original_extension=[".jpg", ".json"],
    )
    file_name = db.original_file_name(file)
    assert file_name == os.path.join(
        legacy_example_dir, "example_filelist", file.path + ".json"
    )

    file = bob.bio.base.database.BioFile(
        4, "data/model4_session1_sample1", "data/model4_session1_sample1"
    )
    with pytest.raises(IOError):
        db.original_file_name(file, False)



import logging

from .. import load_resource
from . import Annotator

logger = logging.getLogger(__name__)


class FailSafe(Annotator):
    """A fail-safe annotator.
    This annotator takes a list of annotator and tries them until you get your
    annotations.
    The annotations of previous annotator is passed to the next one.

    Attributes
    ----------
    annotators : list
        A list of annotators to try
    required_keys : list
        A list of keys that should be available in annotations to stop trying
        different annotators.
    only_required_keys : bool
        If True, the annotations will only contain the ``required_keys``.
    """

    def __init__(
        self, annotators, required_keys, only_required_keys=False, **kwargs
    ):
        super(FailSafe, self).__init__(**kwargs)
        self.annotators = []
        for annotator in annotators:
            if isinstance(annotator, str):
                annotator = load_resource(annotator, "annotator")
            self.annotators.append(annotator)
        self.required_keys = list(required_keys)
        self.only_required_keys = only_required_keys

    def annotate(self, sample, **kwargs):
        if "annotations" not in kwargs or kwargs["annotations"] is None:
            kwargs["annotations"] = {}
        for annotator in self.annotators:
            try:
                annotations = annotator.transform(
                    [sample], **{k: [v] for k, v in kwargs.items()}
                )[0]
            except Exception:
                logger.debug(
                    "The annotator `%s' failed to annotate!",
                    annotator,
                    exc_info=True,
                )
                annotations = None
            if not annotations:
                logger.debug(
                    "Annotator `%s' returned empty annotations.", annotator
                )
            else:
                logger.debug("Annotator `%s' succeeded!", annotator)
            kwargs["annotations"].update(annotations or {})
            # check if we have all the required annotations
            if all(key in kwargs["annotations"] for key in self.required_keys):
                break
        else:  # this else is for the for loop
            # we don't want to return half of the annotations
            kwargs["annotations"] = None
        if self.only_required_keys:
            for key in list(kwargs["annotations"].keys()):
                if key not in self.required_keys:
                    del kwargs["annotations"][key]
        return kwargs["annotations"]

    def transform(self, samples, **kwargs):
        """
        Takes a batch of data and tries annotating them while unsuccessful.

        Tries each annotator given at the creation of FailSafe when the previous
        one fails.

        Each ``kwargs`` value is a list of parameters, with each element of those
        lists corresponding to each element of ``sample_batch`` (for example:
        with ``[s1, s2, ...]`` as ``samples_batch``, ``kwargs['annotations']``
        should contain ``[{<s1_annotations>}, {<s2_annotations>}, ...]``).
        """
        kwargs = translate_kwargs(kwargs, len(samples))
        return [
            self.annotate(sample, **kw) for sample, kw in zip(samples, kwargs)
        ]


def translate_kwargs(kwargs, size):
    new_kwargs = [{}] * size

    if not kwargs:
        return new_kwargs

    for k, value_list in kwargs.items():
        if len(value_list) != size:
            raise ValueError(
                f"Got {value_list} in kwargs which is not of the same length of samples {size}"
            )
        for kw, v in zip(new_kwargs, value_list):
            kw[k] = v

    return new_kwargs



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

from sklearn.base import BaseEstimator, TransformerMixin

from bob.bio.base.preprocessor import Preprocessor


class PreprocessorTransformer(TransformerMixin, BaseEstimator):
    """Scikit learn transformer for :py:class:`bob.bio.base.preprocessor.Preprocessor`.

    Parameters
    ----------
    instance: object
        An instance of `bob.bio.base.preprocessor.Preprocessor`
    """

    def __init__(
        self,
        instance,
        **kwargs,
    ):
        if not isinstance(instance, Preprocessor):
            raise ValueError(
                "`instance` should be an instance of `bob.bio.base.preprocessor.Preprocessor`"
            )

        self.instance = instance
        super().__init__(**kwargs)

    def transform(self, X, annotations=None):
        if annotations is None:
            return [self.instance(data) for data in X]
        else:
            return [
                self.instance(data, annot)
                for data, annot in zip(X, annotations)
            ]

    def _more_tags(self):
        return {
            "requires_fit": False,
            "bob_features_save_fn": self.instance.write_data,
            "bob_features_load_fn": self.instance.read_data,
        }

    def fit(self, X, y=None):
        return self



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

from sklearn.base import BaseEstimator, TransformerMixin

from bob.bio.base.extractor import Extractor

from . import split_X_by_y


class ExtractorTransformer(TransformerMixin, BaseEstimator):
    """Scikit learn transformer for :py:class:`bob.bio.base.extractor.Extractor`.

    Parameters
    ----------
    instance: object
        An instance of :py:class:`bob.bio.base.extractor.Extractor`

    model_path: ``str``
        Model path in case ``instance.requires_training`` is equal to ``True``.
    """

    def __init__(
        self,
        instance,
        model_path=None,
        **kwargs,
    ):
        if not isinstance(instance, Extractor):
            raise ValueError(
                "`instance` should be an instance of `bob.bio.base.extractor.Extractor`"
            )

        if instance.requires_training and (
            model_path is None or model_path == ""
        ):
            raise ValueError(
                f"`model_path` needs to be set if extractor {instance} requires training"
            )

        self.instance = instance
        self.model_path = model_path
        super().__init__(**kwargs)

    def fit(self, X, y=None):
        if not self.instance.requires_training:
            return self

        training_data = X
        if self.instance.split_training_data_by_client:
            training_data = split_X_by_y(X, y)

        self.instance.train(training_data, self.model_path)
        return self

    def transform(self, X, metadata=None):
        if metadata is None:
            return [self.instance(data) for data in X]
        else:
            return [
                self.instance(data, metadata)
                for data, metadata in zip(X, metadata)
            ]

    def _more_tags(self):
        return {
            "requires_fit": self.instance.requires_training,
            "bob_features_save_fn": self.instance.write_feature,
            "bob_features_load_fn": self.instance.read_feature,
        }



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Tiago de Freitas Pereira <tiago.pereira@idiap.ch>


"""Executes only the train part of a biometric pipeline"""

import logging

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option

from bob.pipelines.distributed import VALID_DASK_CLIENT_STRINGS

logger = logging.getLogger(__name__)


EPILOG = """\b

Command line examples\n
-----------------------

$ bob bio pipeline train -vv DATABASE PIPELINE

See the help of the CONFIG argument on top of this help message
for a list of available configurations.

It is possible to provide database and pipeline through a configuration file.
Generate an example configuration file with:

$ bob bio pipeline train --dump-config my_experiment.py

and execute it with:

$ bob bio pipeline train -vv my_experiment.py

my_experiment.py must contain the following elements:

   >>> pipeline = ... # A scikit-learn pipeline wrapped with bob.pipelines' SampleWrapper\n
   >>> database = .... # Biometric Database (class that implements the methods: `background_model_samples`, `references` and `probes`)"
\b"""


@click.command(
    name="train",
    entry_point_group="bob.bio.config",
    cls=ConfigCommand,
    epilog=EPILOG,
)
@click.option(
    "--pipeline",
    "-p",
    required=True,
    entry_point_group="bob.bio.pipeline",
    help="A PipelineSimple or an sklearn.pipeline",
    cls=ResourceOption,
)
@click.option(
    "--database",
    "-d",
    entry_point_group="bob.bio.database",
    required=True,
    help="Biometric Database connector (class that implements the methods: `background_model_samples`, `references` and `probes`)",
    cls=ResourceOption,
)
@click.option(
    "--dask-client",
    "-l",
    entry_point_group="dask.client",
    string_exceptions=VALID_DASK_CLIENT_STRINGS,
    default="single-threaded",
    help="Dask client for the execution of the pipeline.",
    cls=ResourceOption,
)
@click.option(
    "--output",
    "-o",
    show_default=True,
    default="results",
    help="Name of output directory where output files will be saved.",
    cls=ResourceOption,
)
@click.option(
    "--memory",
    "-m",
    is_flag=True,
    help="If set, it will run the experiment keeping all objects on memory with nothing checkpointed. If not set, checkpoints will be saved in `--output`.",
    cls=ResourceOption,
)
@click.option(
    "--checkpoint-dir",
    "-c",
    show_default=True,
    default=None,
    help="Name of output directory where the checkpoints will be saved. In case --memory is not set, checkpoints will be saved in this directory.",
    cls=ResourceOption,
)
@click.option(
    "--dask-partition-size",
    "-s",
    help="If using Dask, this option defines the max size of each dask.bag.partition. "
    "Use this option if the current heuristic that sets this value doesn't suit your experiment. "
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--dask-n-partitions",
    "-n",
    help="If using Dask, this option defines a fixed number of dask.bag.partition for "
    "each set of data. Use this option if the current heuristic that sets this value "
    "doesn't suit your experiment."
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--dask-n-workers",
    "-w",
    help="If using Dask, this option defines the number of workers to start your experiment. "
    "Dask automatically scales up/down the number of workers due to the current load of tasks to be solved. "
    "Use this option if the current amount of workers set to start an experiment doesn't suit you.",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="If set, it will force generate all the checkpoints of an experiment. This option doesn't work if `--memory` is set",
    cls=ResourceOption,
)
@click.option(
    "--no-dask",
    is_flag=True,
    help="If set, it will not use Dask to run the experiment.",
    cls=ResourceOption,
)
@click.option(
    "--split-training",
    is_flag=True,
    help="Splits the training set in partitions and trains the pipeline in multiple steps.",
    cls=ResourceOption,
)
@click.option(
    "--n-splits",
    default=3,
    help="Number of partitions to split the training set in. "
    "Each partition will be trained in a separate step.",
    cls=ResourceOption,
)
@verbosity_option(cls=ResourceOption, logger=logger)
def pipeline_train(
    pipeline,
    database,
    dask_client,
    output,
    memory,
    checkpoint_dir,
    dask_partition_size,
    dask_n_workers,
    dask_n_partitions,
    force,
    no_dask,
    split_training,
    n_splits,
    **kwargs,
):
    """Runs the training part of a biometrics pipeline.

    This pipeline consists only of one component, contrary to the ``simple`` pipeline.
    This component is a scikit-learn ``Pipeline``, where a sequence of transformations
    of the input data is defined.

    The pipeline is trained on the database and the resulting model is saved in the
    output directory.

    It is possible to split the training data in multiple partitions that will be
    used to train the pipeline in multiple steps, helping with big datasets that would
    not fit in memory if trained all at once. Passing the ``--split-training`` option
    will split the training data in ``--n-splits`` partitions and train the pipeline
    sequentially with each partition. The pipeline must support "continuous learning",
    (a call to ``fit`` on an already trained pipeline should continue the training).
    """

    from bob.bio.base.pipelines import execute_pipeline_train

    if no_dask:
        dask_client = None

    checkpoint = not memory

    logger.debug("Executing pipeline training with:")
    logger.debug(f"pipeline: {pipeline}")
    logger.debug(f"database: {database}")

    execute_pipeline_train(
        pipeline=pipeline,
        database=database,
        dask_client=dask_client,
        output=output,
        checkpoint=checkpoint,
        dask_partition_size=dask_partition_size,
        dask_n_partitions=dask_n_partitions,
        dask_n_workers=dask_n_workers,
        checkpoint_dir=checkpoint_dir,
        force=force,
        split_training=split_training,
        n_splits=n_splits,
        **kwargs,
    )

    logger.info(f"Experiment finished ! ({output=})")



import glob
import logging
import os
import pickle
import random

from typing import Optional, Union

import dask.bag
import dask.distributed

from dask.delayed import Delayed
from sklearn.pipeline import Pipeline

from bob.bio.base.pipelines import (
    BioAlgDaskWrapper,
    CSVScoreWriter,
    Database,
    FourColumnsScoreWriter,
    PipelineScoreNorm,
    PipelineSimple,
    TNormScores,
    ZNormScores,
    checkpoint_pipeline_simple,
    dask_bio_pipeline,
    is_biopipeline_checkpointed,
)
from bob.pipelines import (
    DaskWrapper,
    estimator_requires_fit,
    is_instance_nested,
    wrap,
)
from bob.pipelines.distributed import dask_get_partition_size
from bob.pipelines.distributed.sge import SGEMultipleQueuesCluster

logger = logging.getLogger(__name__)


def compute_scores(result, dask_client):
    if isinstance(result, Delayed) or isinstance(result, dask.bag.Bag):
        if dask_client is not None:
            result = result.compute(scheduler=dask_client)
        else:
            logger.warning(
                "`dask_client` not set. Your pipeline will run locally"
            )
            result = result.compute(scheduler="single-threaded")
    return result


def post_process_scores(pipeline, scores, path):
    written_scores = pipeline.write_scores(scores)
    return pipeline.post_process(written_scores, path)


def execute_pipeline_simple(
    pipeline,
    database,
    dask_client,
    groups,
    output,
    write_metadata_scores,
    checkpoint,
    dask_n_partitions,
    dask_partition_size,
    dask_n_workers,
    checkpoint_dir=None,
    force=False,
):
    """
    Function that executes the PipelineSimple.

    This is called when using the ``bob bio pipeline simple``
    command.

    This is also callable from a script without fear of interrupting the running
    Dask instance, allowing chaining multiple experiments while keeping the
    workers alive.

    When using Dask, something to keep in mind is that we want to split our data and
    processing time on multiple workers. There is no recipe to make everything work on
    any system. So if you encounter some balancing error (a few of all the available
    workers actually working while the rest waits, or the scheduler being overloaded
    trying to organize millions of tiny tasks), you can specify ``dask_n_partitions``
    or ``dask_partition_size``.
    The first will try to split any set of data into a number of chunks (ideally, we
    would want one per worker), and the second creates similar-sized partitions in each
    set.
    If the memory on the workers is not sufficient, try reducing the size of the
    partitions (or increasing the number of partitions).

    Parameters
    ----------

    pipeline: Instance of :py:class:`bob.bio.base.pipelines.PipelineSimple`
        A constructed PipelineSimple object.

    database: Instance of :py:class:`bob.bio.base.pipelines.abstract_class.Database`
        A database interface instance

    dask_client: instance of :py:class:`dask.distributed.Client` or ``None``
        A Dask client instance used to run the experiment in parallel on multiple
        machines, or locally.
        Basic configs can be found in ``bob.pipelines.config.distributed``.

    dask_n_partitions: int or None
        Specifies a number of partitions to split the data into.

    dask_partition_size: int or None
        Specifies a data partition size when using dask. Ignored when dask_n_partitions
        is set.

    dask_n_workers: int or None
        Sets the starting number of Dask workers. Does not prevent Dask from requesting
        more or releasing workers depending on load.

    groups: list of str
        Groups of the dataset that will be requested from the database interface.

    output: str
        Path where the scores will be saved.

    write_metadata_scores: bool
        Use the CSVScoreWriter instead of the FourColumnScoreWriter when True.

    checkpoint: bool
        Whether checkpoint files will be created for every step of the pipelines.

    checkpoint_dir: str
        If `checkpoint` is set, this path will be used to save the checkpoints.
        If `None`, the content of `output` will be used.

    force: bool
        If set, it will force generate all the checkpoints of an experiment. This option doesn't work if `--memory` is set
    """
    if not os.path.exists(output):
        os.makedirs(output, exist_ok=True)

    # Setting the `checkpoint_dir`
    if checkpoint_dir is None:
        checkpoint_dir = output
    else:
        os.makedirs(checkpoint_dir, exist_ok=True)

    # Scores are written on `output`
    if write_metadata_scores:
        pipeline.score_writer = CSVScoreWriter(os.path.join(output, "./tmp"))
    else:
        pipeline.score_writer = FourColumnsScoreWriter(
            os.path.join(output, "./tmp")
        )

    # Checkpoint if it's not already checkpointed
    if checkpoint and not is_biopipeline_checkpointed(pipeline):
        hash_fn = database.hash_fn if hasattr(database, "hash_fn") else None
        pipeline = checkpoint_pipeline_simple(
            pipeline, checkpoint_dir, hash_fn=hash_fn, force=force
        )

    # Load the background model samples only if the transformer requires fitting
    if estimator_requires_fit(pipeline.transformer):
        background_model_samples = database.background_model_samples()
    else:
        background_model_samples = []

    for group in groups:
        score_file_name = os.path.join(
            output,
            f"scores-{group}" + (".csv" if write_metadata_scores else ""),
        )
        biometric_references = database.references(group=group)
        probes = database.probes(group=group)

        # If there's no data to be processed, continue
        if len(biometric_references) == 0 or len(probes) == 0:
            logger.warning(
                f"Current dataset ({database}) does not have `{group}` set. The experiment will not be executed."
            )
            continue

        if dask_client is not None and not is_instance_nested(
            pipeline.biometric_algorithm,
            "biometric_algorithm",
            BioAlgDaskWrapper,
        ):
            # Scaling up
            if dask_n_workers is not None and not isinstance(dask_client, str):
                dask_client.cluster.scale(dask_n_workers)

            # Data partitioning.
            # - Too many small partitions: the scheduler takes more time scheduling
            #   than the computations.
            # - Too few big partitions: We don't use all the available workers and thus
            #   run slower.
            if dask_partition_size is not None:
                logger.debug(
                    f"Splitting data with fixed size partitions: {dask_partition_size}."
                )
                pipeline = dask_bio_pipeline(
                    pipeline,
                    partition_size=dask_partition_size,
                )
            elif dask_n_partitions is not None or dask_n_workers is not None:
                # Divide each Set in a user-defined number of partitions
                n_partitions = dask_n_partitions or dask_n_workers
                logger.debug(
                    f"Splitting data with fixed number of partitions: {n_partitions}."
                )
                pipeline = dask_bio_pipeline(pipeline, npartitions=n_partitions)
            else:
                # Split in max_jobs partitions or revert to the default behavior of
                # dask.Bag from_sequence: partition_size = 100
                n_jobs = None
                if not isinstance(dask_client, str) and isinstance(
                    dask_client.cluster, SGEMultipleQueuesCluster
                ):
                    logger.debug(
                        "Splitting data according to the number of available workers."
                    )
                    n_jobs = dask_client.cluster.sge_job_spec["default"][
                        "max_jobs"
                    ]
                    logger.debug(f"{n_jobs} partitions will be created.")
                pipeline = dask_bio_pipeline(pipeline, npartitions=n_jobs)

        logger.info(f"Running the PipelineSimple for group {group}")
        score_all_vs_all = (
            database.score_all_vs_all
            if hasattr(database, "score_all_vs_all")
            else False
        )

        result = pipeline(
            background_model_samples,
            biometric_references,
            probes,
            score_all_vs_all=score_all_vs_all,
        )

        post_processed_scores = post_process_scores(
            pipeline, result, score_file_name
        )
        compute_scores(post_processed_scores, dask_client)


def execute_pipeline_score_norm(
    pipeline,
    database,
    dask_client,
    groups,
    output,
    write_metadata_scores,
    checkpoint,
    dask_partition_size,
    dask_n_workers,
    checkpoint_dir=None,
    top_norm=False,
    top_norm_score_fraction=0.8,
    score_normalization_type="znorm",
    force=False,
):
    """
    Function that extends the capabilities of the PipelineSimple to run score normalization.

    This is called when using the ``bob bio pipeline score-norm`` command.

    This is also callable from a script without fear of interrupting the running
    Dask instance, allowing chaining multiple experiments while keeping the
    workers alive.

    Parameters
    ----------

    pipeline: Instance of :py:class:`bob.bio.base.pipelines.PipelineSimple`
        A constructed PipelineSimple object.

    database: Instance of :py:class:`bob.bio.base.pipelines.abstract_class.Database`
        A database interface instance

    dask_client: instance of :py:class:`dask.distributed.Client` or ``None``
        A Dask client instance used to run the experiment in parallel on multiple machines, or locally. Basic configs can be found in ``bob.pipelines.config.distributed``.

    groups: list of str
        Groups of the dataset that will be requested from the database interface.

    output: str
        Path where the results and checkpoints will be saved to.

    write_metadata_scores: bool
        Use the CSVScoreWriter instead of the FourColumnScoreWriter when True.

    checkpoint: bool
        Whether checkpoint files will be created for every step of the pipelines.

    dask_partition_size: int
        If using Dask, this option defines the size of each dask.bag.partition. Use this option if the current heuristic that sets this value doesn't suit your experiment. (https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).

    dask_n_workers: int
        If using Dask, this option defines the number of workers to start your experiment. Dask automatically scales up/down the number of workers due to the current load of tasks to be solved. Use this option if the current amount of workers set to start an experiment doesn't suit you.

    top_norm: bool

    top_norm_score_fraction: float
        Sets the percentage of samples used for t-norm and z-norm. Sometimes you don't want to use all the t/z samples for normalization

    checkpoint_dir: str
        If `checkpoint` is set, this path will be used to save the checkpoints.
        If `None`, the content of `output` will be used.

    """

    if not os.path.exists(output):
        os.makedirs(output, exist_ok=True)

    # Setting the `checkpoint_dir`
    if checkpoint_dir is None:
        checkpoint_dir = output
    else:
        os.makedirs(checkpoint_dir, exist_ok=True)

    # Scores are written on `output`
    if write_metadata_scores:
        pipeline.score_writer = CSVScoreWriter(os.path.join(output, "./tmp"))
    else:
        pipeline.score_writer = FourColumnsScoreWriter(
            os.path.join(output, "./tmp")
        )

    # Check if it's already checkpointed
    if checkpoint and not is_biopipeline_checkpointed(pipeline):
        pipeline = checkpoint_pipeline_simple(
            pipeline, checkpoint_dir, force=force
        )

    # PICKING THE TYPE OF POST-PROCESSING
    if score_normalization_type == "znorm":
        post_processor = ZNormScores(
            top_norm=top_norm,
            top_norm_score_fraction=top_norm_score_fraction,
        )
    elif score_normalization_type == "tnorm":
        post_processor = TNormScores(
            top_norm=top_norm,
            top_norm_score_fraction=top_norm_score_fraction,
        )
    else:
        raise ValueError(
            f"score_normalization_type {score_normalization_type} is not valid"
        )

    if checkpoint and not is_biopipeline_checkpointed(post_processor):
        score_stats_path = os.path.join(
            checkpoint_dir,
            f"{score_normalization_type}-scores",
            "norm",
            "stats",
        )
        # we cannot checkpoint "features" because sample.keys are not unique.
        post_processor = wrap(
            ["checkpoint"],
            post_processor,
            model_path=score_stats_path,
            force=force,
        )

    pipeline = PipelineScoreNorm(pipeline, post_processor)

    background_model_samples = database.background_model_samples()

    # treferences = database.treferences(proportion=ztnorm_cohort_proportion)
    for group in groups:
        # Changing the score normalization stats file name as a function of the group
        if checkpoint and not is_biopipeline_checkpointed(post_processor):
            post_processor.model_path = f"{score_stats_path}_{group}.pkl"

        if score_normalization_type == "znorm":
            score_normalization_samples = database.zprobes(group=group)
        elif score_normalization_type == "tnorm":
            score_normalization_samples = database.treferences()

        score_file_name = os.path.join(output, f"scores-{group}")

        biometric_references = database.references(group=group)
        probes = database.probes(group=group)

        # If there's no data to be processed, continue
        if len(biometric_references) == 0 or len(probes) == 0:
            logger.warning(
                f"Current dataset ({database}) does not have `{group}` set. The experiment will not be executed."
            )
            continue

        if dask_client is not None and not is_instance_nested(
            pipeline.biometric_algorithm,
            "biometric_algorithm",
            BioAlgDaskWrapper,
        ):
            # Scaling up
            if dask_n_workers is not None and not isinstance(dask_client, str):
                dask_client.cluster.scale(dask_n_workers)

            n_objects = max(
                len(background_model_samples),
                len(biometric_references),
                len(probes),
            )
            partition_size = None
            if not isinstance(dask_client, str):
                partition_size = dask_get_partition_size(
                    dask_client.cluster, n_objects
                )
            if dask_partition_size is not None:
                partition_size = dask_partition_size

            pipeline = dask_bio_pipeline(
                pipeline,
                partition_size=partition_size,
            )

        logger.info(f"Running PipelineSimple for group {group}")
        score_all_vs_all = (
            database.score_all_vs_all
            if hasattr(database, "score_all_vs_all")
            else False
        )

        (
            raw_scores,
            score_normed_scores,
        ) = pipeline(
            background_model_samples,
            biometric_references,
            probes,
            score_normalization_samples,
            score_all_vs_all=score_all_vs_all,
        )

        def _build_filename(score_file_name, suffix):
            return os.path.join(score_file_name, suffix)

        # Running RAW_SCORES

        raw_scores = post_process_scores(
            pipeline,
            raw_scores,
            _build_filename(score_file_name, "raw_scores.csv"),
        )
        _ = compute_scores(raw_scores, dask_client)

        # Z-SCORES
        score_normed_scores = post_process_scores(
            pipeline,
            score_normed_scores,
            _build_filename(score_file_name, f"{score_normalization_type}.csv"),
        )
        _ = compute_scores(score_normed_scores, dask_client)

        # T-SCORES
        """
        t_normed_scores = post_process_scores(
            pipeline,
            t_normed_scores,
            _build_filename(score_file_name, "t_normed_scores.csv"),
        )
        _ = compute_scores(t_normed_scores, dask_client)

        # S-SCORES
        s_normed_scores = post_process_scores(
            pipeline,
            s_normed_scores,
            _build_filename(score_file_name, "s_normed_scores.csv"),
        )
        _ = compute_scores(s_normed_scores, dask_client)

        # ZT-SCORES
        zt_normed_scores = post_process_scores(
            pipeline,
            zt_normed_scores,
            _build_filename(score_file_name, "zt_normed_scores.csv"),
        )
        _ = compute_scores(zt_normed_scores, dask_client)
        """


def execute_pipeline_train(
    pipeline: Union[PipelineSimple, Pipeline],
    database: Database,
    dask_client: Optional[dask.distributed.Client] = None,
    output: str = "./results",
    checkpoint: bool = True,
    dask_n_partitions: Optional[int] = None,
    dask_partition_size: Optional[int] = None,
    dask_n_workers: Optional[int] = None,
    checkpoint_dir: Optional[str] = None,
    force: bool = False,
    split_training: bool = False,
    n_splits: int = 3,
    **kwargs,
):
    """Executes only the training part of a pipeline.

    When running from a script, use this function instead of the click command in
    ``bob.bio.base.script.pipeline_train``.

    Parameters
    ----------

    pipeline:
        A constructed ``PipelineSimple`` object (the ``transformer`` will be extracted
        for training), or an ``sklearn.Pipeline``.

    database:
        A database interface instance

    dask_client:
        A Dask client instance used to run the experiment in parallel on multiple
        machines, or locally.
        Basic configs can be found in ``bob.pipelines.config.distributed``.

    dask_n_partitions:
        Specifies a number of partitions to split the data into.

    dask_partition_size:
        Specifies a data partition size when using dask. Ignored when dask_n_partitions
        is set.

    dask_n_workers:
        Sets the starting number of Dask workers. Does not prevent Dask from requesting
        more or releasing workers depending on load.

    output:
        Path where the scores will be saved.

    checkpoint:
        Whether checkpoint files will be created for every step of the pipelines.

    checkpoint_dir:
        If `checkpoint` is set, this path will be used to save the checkpoints.
        If `None`, the content of `output` will be used.

    force:
        If set, it will force generate all the checkpoints of an experiment. This option doesn't work if `--memory` is set

    split_training:
        If set, the background model will be trained on multiple partitions of the data.

    n_splits:
        Number of splits to use when splitting the data.
    """

    logger.debug(f"Unused arguments: {kwargs=}")
    if not os.path.exists(output):
        os.makedirs(output, exist_ok=True)

    # Setting the `checkpoint_dir`
    if checkpoint_dir is None:
        checkpoint_dir = output
    else:
        os.makedirs(checkpoint_dir, exist_ok=True)

    if isinstance(pipeline, PipelineSimple):
        pipeline = pipeline.transformer

    # Checkpoint (only features, not the model)
    if checkpoint:
        hash_fn = database.hash_fn if hasattr(database, "hash_fn") else None
        wrap(
            ["checkpoint"],
            pipeline,
            features_dir=checkpoint_dir,
            model_path=None,
            hash_fn=hash_fn,
            force=force,
        )

    if not estimator_requires_fit(pipeline):
        raise ValueError(
            "Estimator does not require fitting. No training necessary."
        )

    background_model_samples = database.background_model_samples()

    if dask_client is not None:
        # Scaling up
        if dask_n_workers is not None and not isinstance(dask_client, str):
            dask_client.cluster.scale(dask_n_workers)

        if dask_partition_size is not None:
            logger.debug(
                f"Splitting data with fixed size partitions: {dask_partition_size}."
            )
            pipeline = wrap(
                ["dask"], pipeline, partition_size=dask_partition_size
            )
        elif dask_n_partitions is not None or dask_n_workers is not None:
            # Divide each Set in a user-defined number of partitions
            n_partitions = dask_n_partitions or dask_n_workers
            logger.debug(
                f"Splitting data with fixed number of partitions: {n_partitions}."
            )
            pipeline = wrap(["dask"], pipeline, npartitions=n_partitions)
        else:
            # Split in max_jobs partitions or revert to the default behavior of
            # dask.Bag from_sequence: partition_size = 100
            n_jobs = None
            if not isinstance(dask_client, str) and isinstance(
                dask_client.cluster, SGEMultipleQueuesCluster
            ):
                logger.debug(
                    "Splitting data according to the number of available workers."
                )
                n_jobs = dask_client.cluster.sge_job_spec["default"]["max_jobs"]
                logger.debug(f"{n_jobs} partitions will be created.")
            pipeline = wrap(["dask"], pipeline, npartitions=n_jobs)

    logger.info("Running the pipeline training")
    if split_training:
        start_step = -1
        # Look at step files, and assess if we can load the last one
        for step_file in glob.glob(
            os.path.join(output, "train_pipeline_step_*.pkl")
        ):
            to_rem = os.path.join(output, "train_pipeline_step_")
            file_step = int(step_file.replace(to_rem, "").replace(".pkl", ""))
            start_step = max(start_step, file_step)
        if start_step > -1:
            logger.debug("Found pipeline training step. Loading it.")
            last_step_file = os.path.join(
                output, f"train_pipeline_step_{start_step}.pkl"
            )
            with open(last_step_file, "rb") as start_file:
                pipeline = pickle.load(start_file)
        start_step += 1  # Loaded step is i -> training starts a i+1
        logger.info(f"Starting from training step {start_step}")

        random.seed(0)
        random.shuffle(background_model_samples)

        for partition_i in range(start_step, n_splits):
            logger.info(
                f"Training with partition {partition_i} ({partition_i+1}/{n_splits})"
            )
            start = len(background_model_samples) // n_splits * partition_i
            end = len(background_model_samples) // n_splits * (partition_i + 1)
            _ = pipeline.fit(background_model_samples[start:end])
            step_path = os.path.join(
                output, f"train_pipeline_step_{partition_i}.pkl"
            )
            with open(step_path, "wb") as f:
                pickle.dump(pipeline, f)
    else:
        _ = pipeline.fit(background_model_samples)

    # Save each fitted transformer
    for transformer_name, transformer in pipeline.steps:
        if transformer._get_tags()["requires_fit"]:
            if isinstance(transformer, DaskWrapper):
                transformer = transformer.estimator
            step_path = os.path.join(output, f"{transformer_name}.pkl")
            with open(step_path, "wb") as f:
                pickle.dump(transformer, f)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

import collections
import json
import logging
import os

from bob.bio.base.database.utils import search_and_open

logger = logging.getLogger(__name__)

_idiap_annotations = {
    1: "reyeo",
    2: "reyet",
    3: "reyep",
    4: "reyeb",
    5: "reyei",
    6: "leyei",
    7: "leyet",
    8: "leyep",
    9: "leyeb",
    10: "leyeo",
    11: "rbrowo",
    12: "rbrowi",
    13: "lbrowi",
    14: "lbrowo",
    15: "noser",
    16: "noset",
    17: "nosel",
    18: "mouthr",
    19: "moutht",
    20: "mouthb",
    21: "mouthl",
    22: "chin",
}


def read_annotation_file(file_name, annotation_type):
    """This function provides default functionality to read annotation files.

    Parameters
    ----------
    file_name : str
        The full path of the annotation file to read. The path can also be like
        ``base_path:relative_path`` where the base_path can be both a directory or
        a tarball. This allows you to read annotations from inside a tarball.
    annotation_type : str
        The type of the annotation file that should be read. The following
        annotation_types are supported:

          * ``eyecenter``: The file contains a single row with four entries:
            ``re_x re_y le_x le_y``
          * ``named``: The file contains named annotations, one per line, e.g.:
            ``reye re_x re_y`` or ``pose 25.7``
          * ``idiap``: The file contains enumerated annotations, one per line,
            e.g.: ``1 key1_x key1_y``, and maybe some additional annotations like
            gender, age, ...
          * ``json``: The file contains annotations of any format, dumped in a
            text json file.

    Returns
    -------
    dict
        A python dictionary with the keypoint name as key and the
        position ``(y,x)`` as value, and maybe some additional annotations.

    Raises
    ------
    IOError
        If the annotation file is not found.
    ValueError
        If the annotation type is not known.
    """
    if not file_name:
        return None

    if ":" in file_name:
        base_path, tail = file_name.split(":", maxsplit=1)
        f = search_and_open(search_pattern=tail, base_dir=base_path)
    else:
        if not os.path.exists(file_name):
            raise IOError("The annotation file '%s' was not found" % file_name)
        f = open(file_name)

    annotations = {}

    try:
        if str(annotation_type) == "eyecenter":
            # only the eye positions are written, all are in the first row
            line = f.readline()
            positions = line.split()
            assert len(positions) == 4
            annotations["reye"] = (float(positions[1]), float(positions[0]))
            annotations["leye"] = (float(positions[3]), float(positions[2]))

        elif str(annotation_type) == "named":
            # multiple lines, no header line, each line contains annotation and
            # position or single value annotation
            for line in f:
                positions = line.split()
                if len(positions) == 3:
                    annotations[positions[0]] = (
                        float(positions[2]),
                        float(positions[1]),
                    )
                elif len(positions) == 2:
                    annotations[positions[0]] = float(positions[1])
                else:
                    logger.error(
                        "Could not interpret line '%s' in annotation file '%s'",
                        line,
                        file_name,
                    )

        elif str(annotation_type) == "idiap":
            # Idiap format: multiple lines, no header, each line contains an integral
            # keypoint identifier, or other identifier like 'gender', 'age',...
            for line in f:
                positions = line.rstrip().split()
                if positions:
                    if positions[0].isdigit():
                        # position field
                        assert len(positions) == 3
                        id = int(positions[0])
                        annotations[_idiap_annotations[id]] = (
                            float(positions[2]),
                            float(positions[1]),
                        )
                    else:
                        # another field, we take the first entry as key and the rest as
                        # values
                        annotations[positions[0]] = positions[1:]
            # finally, we add the eye center coordinates as the center between the
            # eye corners; the annotations 3 and 8 are the pupils...
            if "reyeo" in annotations and "reyei" in annotations:
                annotations["reye"] = (
                    (annotations["reyeo"][0] + annotations["reyei"][0]) / 2.0,
                    (annotations["reyeo"][1] + annotations["reyei"][1]) / 2.0,
                )
            if "leyeo" in annotations and "leyei" in annotations:
                annotations["leye"] = (
                    (annotations["leyeo"][0] + annotations["leyei"][0]) / 2.0,
                    (annotations["leyeo"][1] + annotations["leyei"][1]) / 2.0,
                )

        elif str(annotation_type) == "json":
            annotations = json.load(
                f, object_pairs_hook=collections.OrderedDict
            )
        else:
            raise ValueError(
                "The given annotation type '%s' is not known, choose one of ('eyecenter', 'named', 'idiap')"
                % annotation_type
            )
    finally:
        f.close()

    if (
        annotations is not None
        and "leye" in annotations
        and "reye" in annotations
        and annotations["leye"][1] < annotations["reye"][1]
    ):
        logger.warn(
            "The eye annotations in file '%s' might be exchanged!" % file_name
        )

    return annotations



import numpy as np

from sklearn.preprocessing import OrdinalEncoder


class ReferenceIdEncoder(OrdinalEncoder):
    """An OrdinalEncoder that can converts subject_id strings to integers.
    This is used to prepare labels used in training supervised transformers like
    the ISV algorithm.
    """

    # Default values of init args are different from the base class
    def __init__(
        self,
        *,
        categories="auto",
        dtype=int,
        handle_unknown="use_encoded_value",
        unknown_value=-1,
        **kwargs,
    ):
        super().__init__(
            categories=categories,
            dtype=dtype,
            handle_unknown=handle_unknown,
            unknown_value=unknown_value,
            **kwargs,
        )

    def fit(self, X, y=None):
        # X is a SampleBatch or list of subject_id strings
        # we want a 2d array of shape (N, 1)
        X = np.asarray(X).reshape((-1, 1))
        return super().fit(X)

    def transform(self, X):
        X = np.asarray(X).reshape((-1, 1))
        # we output a flat array instead
        return super().transform(X).flatten()

    def _more_tags(self):
        return {
            "bob_input": "subject_id",
            "bob_output": "subject_id_int",
        }
